# $Revision: 32 $
# Copyright 2006-2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from c_numpy cimport npy_intp

cdef extern from "fwd_bwd.h":
  ctypedef double (*inner_funcptr)(double *, npy_intp *, double *, double *,
                                   npy_intp *, npy_intp, npy_intp, npy_intp,
                                   npy_intp, npy_intp, npy_intp)

  cdef double fwd_inner(double *, npy_intp *, double *, double *, npy_intp *,
                        npy_intp, npy_intp, npy_intp, npy_intp, npy_intp,
                        npy_intp)
  cdef double bwd_inner(double *, npy_intp *, double *, double *, npy_intp *,
                        npy_intp, npy_intp, npy_intp, npy_intp, npy_intp,
                        npy_intp)



#!/usr/bin/env python
from __future__ import division, with_statement

"""
make_species_info: DESCRIPTION
"""

__version__ = "$Revision: 256 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

INPUT_FILENAME = "all.fastacomposition"
OUTPUT_FILENAME = "../../sunflower/data/species.pck"

from cPickle import dump
import sys

from sunflower.hmm import PICKLE_PROTOCOL
from sunflower.recompose import cols2logprobs_emission

def make_species_info(filename=INPUT_FILENAME):
    info = {}

    with open(filename) as infile:
        for line in infile:
            cols = line.rstrip().split()

            species = cols[0]
            resourcename = cols[1]
            count_cols = cols[2:]

            logprobs_emission = cols2logprobs_emission(count_cols)

            info[species] = (resourcename, logprobs_emission)

    with open(OUTPUT_FILENAME, "wb") as outfile:
        dump(info, outfile, PICKLE_PROTOCOL)

def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... [FILE]"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)

    options, args = parser.parse_args(args)

    if not len(args) <= 1:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    return make_species_info(*args)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

"""
h5cat: h5cat two HDF5 files together without
overwriting nodes or throwing errors
"""

__version__ = "$Revision: 413 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from errno import EISDIR
from os import strerror
import sys

from path import path
from tables import NodeError

from ._utils import DESC, progress
from .hdf5 import h5open

def concatenate_file(src_file, dst_file):
    for src_node in src_file.root._f_walkNodes():
        group_pathname = src_node._v_parent._v_pathname
        dst_group = dst_file.getNode(group_pathname)

        try:
            src_node._f_copy(dst_group)
        except NodeError:
            pass

def h5cat(src_filenames, dst_filename, recursive=False, ignore_errors=False):
    with h5open(dst_filename, "w", DESC) as dst_file:
        for src_filename in src_filenames:
            progress(src_filename)

            src_filepath = path(src_filename)
            if src_filepath.isdir():
                if recursive:
                    src_filenames.extend(src_filepath.listdir())
                    continue
                else:
                    raise IOError(EISDIR, strerror(EISDIR))

            try:
                with h5open(src_filename) as src_file:
                    concatenate_file(src_file, dst_file)
            except IOError, err:
                if ignore_errors:
                    err_msg = "ignored %s: %s" % (err.__class__.__name__, err)
                    print >>sys.stderr, err_msg
                else:
                    raise

def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... SRC... DST"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--recursive", action="store_true",
                      help="act recursively on directories")
    parser.add_option("--ignore-errors", action="store_true",
                      help="continue to the next file on error instead of "
                      "terminating")

    options, args = parser.parse_args(args)

    if not len(args) >= 2:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    h5cat(args[:-1], args[-1], recursive=options.recursive,
          ignore_errors=options.ignore_errors)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

"""
priors: alters the prior probabilities of transitions from the
silent state to various transcription factor "petals"
"""

__version__ = "$Revision: 428 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from collections import defaultdict
import sys

from numpy import array, float64, log as alog
from math import log, exp

from ._utils import DESC
from .hmm import (SPECIAL_STATE_NAMES, OFFSET_UNBOUND, OFFSET_SILENT,
                  TF_START_FINAL_SUFFIXES,
                  get_state_names, get_tf_info, get_transitions, 
                  adjust_transitions, load, save, dump)

def load_priors(filename):

    priors = {}

    ## read the file - must be in the form TF prior
    with open(filename) as infile:
        for line in infile:
            tokens = line.split()
            if len(tokens) == 2:
                priors[tokens[0]] = float(tokens[1])

    ## check that the unbound prior was set
    assert SPECIAL_STATE_NAMES[OFFSET_UNBOUND] in priors

    ## check that priors sum to 1 - if not, scale them
    total = sum(priors.values())
    if total != 1.0:
        # XXX: vectorize
        for key in priors.keys():
            priors[key] = priors[key] / total

    return priors

def update_priors(model, priors):

    state_names = get_state_names(model)
    (tf_start_offsets, tf_end_offsets, revcom) = get_tf_info(model)
    logprobs_a = get_transitions(model)

    ## map tf names to start offsets (works for both fwdonly and revcom)
    tf_to_start_offsets = {}
    for tf_start_offset in tf_start_offsets:
        
        last_dot = state_names[tf_start_offset].rindex(".")
        tf_name = state_names[tf_start_offset][:last_dot]

        if tf_name not in tf_to_start_offsets:
            tf_to_start_offsets[tf_name] = list()

        tf_to_start_offsets[tf_name].append(tf_start_offset)

    for tf_name in priors:

        ## unbound is a special case
        if tf_name == SPECIAL_STATE_NAMES[OFFSET_UNBOUND]:
            logprobs_a[OFFSET_SILENT, OFFSET_UNBOUND] = log(priors[tf_name])

        else:
            for tf_start_offset in tf_to_start_offsets[tf_name]:

                cell = (OFFSET_SILENT, tf_start_offset)

                if revcom:
                    logprobs_a[cell] = log(priors[tf_name] / 2)
                else:
                    logprobs_a[cell] = log(priors[tf_name])

    return logprobs_a

def priors(priors_filename, infilename, outfilename, resource=False):

    model = load(infilename, resource)
    priors = load_priors(priors_filename)

    logprobs_transition = update_priors(model, priors)

    adjust_transitions(model, logprobs_transition)
    model["desc"] = "; ".join([model["desc"], DESC])

    save(outfilename, model)

def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... PRIORS IN-SFL OUT-SFL"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-R", "--resource", action="store_true",
                      help="get SFLFILE from a resource that comes with the "
                      "sunflower distribution instead of a file")

    options, args = parser.parse_args(args)

    if not len(args) == 3:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):

    options, args = parse_options(args)

    return priors(resource=options.resource, *args)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

__version__ = "$Revision: 446 $"

"""
simulator: DESCRIPTION
"""

# Copyright 2006-2009 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

### imports

from collections import defaultdict
from itertools import count, izip
from math import log
import re
import sys

from numpy import (allclose, array, empty, empty_like, exp as aexp,
                   isfinite, index_exp, intp, float64, ndarray, zeros)
from textinput import open as _open

from ._fwd_bwd import fwd, bwd
from ._poly import chunk
from ._utils import FastaIterator, enum
from .driver import defline_identifier, load as driver_load
from .hmm import (ALPHABET, ALPHABET_INT, LEN_ALPHABET, get_params,
                  load as hmm_load)

### constants

ALPHABET_ENUM = enum(ALPHABET)
LN_2 = log(2)

### accumulators

def accum_sum_relent_ref_mut(probs_ref, logprobs_ref, logprobs_mut,
                             mask_nonzero_prob_cols):
    logprobs_ref_masked = logprobs_ref[mask_nonzero_prob_cols]
    logprobs_mut_masked = logprobs_mut[mask_nonzero_prob_cols]

    probs_ref_masked = probs_ref[mask_nonzero_prob_cols]

    # relative entropy: Durbin and co-workers p. 308
    # calculate log_2 [P(ref)/P(mut)] from ln P(ref) and ln P(mut):
    ref_mut_bits = (logprobs_ref_masked - logprobs_mut_masked)/LN_2

    # relative entropy is for one probability distribution but we are
    # summing across a distribution for each nucleotide

    # if you run this on the 5' end of a sequence, then you will get
    # NaNs due to the zero probability areas subtracting in log
    # space.
    return (probs_ref_masked * ref_mut_bits).sum()

### other functions

def seq2int(seq):
    res = [-1] + [ALPHABET_ENUM[char] for char in seq.upper()]
    return intp(res)

def match_names(pattern, names):
    """
    returns indexes_match, names_match
    """
    regex = re.compile(pattern)
    matchfunc = regex.search

    zipped = ((index, name) for index, name in enumerate(names)
              if matchfunc(name))

    # unzip and arrayize
    return tuple(array(x) for x in zip(*zipped))

def load_seqrecords(seq_filename, include, exclude):
    res = []

    if seq_filename is None:
        return res

    excluded_ids = set(exclude)
    included_ids = set(include)

    with _open(seq_filename) as seq_file:
        for defline, seq in FastaIterator(seq_file):
            identifier = defline_identifier(defline)
            if (not included_ids) or (identifier in included_ids):
                if identifier not in excluded_ids:
                    res.append((defline, seq))

    if not res:
        print >>sys.stderr, "no matching sequences"
        sys.exit(2)

    return res

def _assert_fwd_bwd_results(f, b, logprob_tf, logprob_tb):
    assert allclose(logprob_tf, logprob_tb)

    # we'll run out of memory if we don't specify output arrays:
    # XXX: (a numexpr or blitz approach would be better, but require
    # yet another prereq)
    arr = f[1:, 1:] + b[1:, 1:]
    arr -= logprob_tf
    aexp(arr, arr)

    assert allclose(arr.sum(-1), 1.0)

def fwd_bwd(f, b, x, (a, e, cf, cb), cells_keep=index_exp[:], start=0):
    """
    start allows you to avoid repeating all of the fwd and bwd calculations
    after a mutation
    """
    logprob_tf = fwd(f, x, a, e, cf, start)
    logprob_tb = bwd(b, x, a, e, cb, start)

    if __debug__:
        _assert_fwd_bwd_results(f, b, logprob_tf, logprob_tb)

    logprobs = f[cells_keep] + b[cells_keep] - logprob_tf

    return aexp(logprobs)

def fwd_bwd_log(f, b, x, (a, e, cf, cb), start=0):
    """
    returns log values instead of exp'd
    doesn't support cells_keep
    """
    logprob_tf = fwd(f, x, a, e, cf, start)
    logprob_tb = bwd(b, x, a, e, cb, start)

    if __debug__:
        _assert_fwd_bwd_results(f, b, logprob_tf, logprob_tb)

    return f + b - logprob_tf

# per-sequence
def simulate_mutations(f, b, x, model_params, rows_keep, groupings):
    # slice setup
    assert rows_keep.step is None
    if not rows_keep.start: # if it is 0 or None
        # XXX: I don't think this ever happens
        rows_keep = slice(1, rows_keep.stop)

    x_keep_range = range(len(x))[rows_keep]
    x_keep = x[rows_keep]

    # array setup

    # res: (position, nucleotide)
    # must be zeros, since some spaces are never set
    shape = (len(x_keep_range), LEN_ALPHABET)
    res = zeros(shape, dtype=float64)

    # initial calculation
    logprobs_ref = fwd_bwd_log(f, b, x, model_params)
    probs_ref = aexp(logprobs_ref)

    # mask out sequence columns with 0s (NINF in log space) to avoid
    # dividing by zero in calculating relative entropy
    mask_nonzero_prob_cols = isfinite(logprobs_ref.sum(1))

    x_mut = x.copy()

    # will have to be refreshed after each mutation
    f_mut = empty_like(f)
    i_first_minus1 = x_keep_range[0]-1
    f_mut[:i_first_minus1] = f[:i_first_minus1]

    # b is re-used over and over again: only gets erased one at a time in
    # the right direction.

    # iteration: *_mut
    # limit to rows_keep:
    for res_index, i, x_i in izip(count(), x_keep_range, x_keep):
        f_mut[i-1] = f[i-1]

        for x_i_mut in ALPHABET_INT:
            if x_i_mut == x_i:
                continue

            x_mut[i] = x_i_mut

            logprobs_mut = fwd_bwd_log(f_mut, b, x_mut, model_params, i)

            res[res_index, x_i_mut] = \
                accum_sum_relent_ref_mut(probs_ref, logprobs_ref, logprobs_mut,
                                         mask_nonzero_prob_cols)

        # return to reference state to prepare for next round
        x_mut[i] = x_i

    return res

def regroup_probs(probs, groupings):
    # XXXopt: try adding an extra row of zeros and filling in
    # groupings to be of a homogeneous shape with -1
    shape = (probs.shape[0], len(groupings))
    res = empty(shape, probs.dtype)

    for grouping_index, grouping in enumerate(groupings):
        res[:, grouping_index] = probs[:, grouping].sum(1)

    return res

def simulate_diff(f, b, x, model_params, rows_keep, groupings):
    x, x_mut = x

    # slice is not supported--it does not make sense in this context
    # the slice is normally used to specify the rows on which
    # mutations occur, but they can occur anywhere here. It is up to
    # the user to limit the effects of mutations near the edges
    # XXX: move out of this func
    assert (rows_keep.start == 1 and rows_keep.stop is None
            and rows_keep.step is None)

    # initial calculation
    logprobs_ref = fwd_bwd_log(f, b, x, model_params)
    probs_ref = aexp(logprobs_ref)

    # reuses f, b
    logprobs_mut = fwd_bwd_log(f, b, x_mut, model_params)

    # mask out sequence columns with 0s (NINF in log space) to avoid
    # dividing by zero in calculating relative entropy
    mask_nonzero_prob_cols = isfinite(logprobs_ref.sum(1))

    return accum_sum_relent_ref_mut(probs_ref, logprobs_ref, logprobs_mut,
                                    mask_nonzero_prob_cols)

def simulate_reference(f, b, x, model_params, cells_keep, groupings):
    # fwd_bwd() limits to cells_keep:
    probs = fwd_bwd(f, b, x, model_params, cells_keep)

    if groupings:
        return regroup_probs(probs, groupings)
    else:
        return probs

def simulate_seq(simulate_func, driver, model_params, num_states, seqrecord,
                 cells_keep, groupings):
    defline, seq = seqrecord

    if isinstance(seq, ndarray):
        # overloaded with two sequences
        assert len(seq) == 2

        x = [seq2int(seq_inner) for seq_inner in seq]
        f_shape = (len(x[0]), num_states)
        assert f_shape[0] == len(x[1])
    else:
        # default case
        x = seq2int(seq)
        f_shape = (len(x), num_states)

    f = empty(f_shape, dtype=float64)
    b = empty(f_shape, dtype=float64)

    outarray = simulate_func(f, b, x, model_params, cells_keep, groupings)

    # store the data
    driver(outarray, seqrecord)

def correct_limit_slice(limit_slice):
    limit_start = limit_slice.start
    limit_stop = limit_slice.stop

    # slice setup
    if not limit_start:
        limit_start = 1
    elif limit_start > 0: # negative offsets unchanged
        limit_start += 1

    if limit_stop >= 0:
        limit_stop += 1

    return slice(limit_start, limit_stop, limit_slice.step)

def group_stems(names):
    """
    returns a defaultdict:

    keys: group stem names
    values: matching indexes
    """
    res = defaultdict(list)

    for index, name in enumerate(names):
        name = name.rpartition(".")[0] or name
        res[name].append(index)

    return res

def make_groupings(colnames):
    groups = group_stems(colnames)

    groupings = [] # a list of lists
    colnames = [] # a list of strs (convert to an array of strs)

    # ensures consistent collinear order, unlike a dict
    for colname, col_indexes in groups.iteritems():
        groupings.append(col_indexes)
        colnames.append(colname)

    return groupings, array(colnames)

# XXX: too many args
def simulator(hmm_filename, seq_filename, args, driver_class=None,
              simulate_func=simulate_reference,
              keep_pattern=None, include=[],
              exclude=[], resource=False, start_default=None, write_seq=True,
              limit_slice=None, total_stems=False, seq2_filename=None):
    model = hmm_load(hmm_filename, resource)
    num_states = model["num_states"]

    if __debug__:
        alphabet = model["alphabet"]
        emitting_states = model["emitting_states"]

        assert alphabet == ALPHABET
        assert emitting_states == range(1, num_states)

    state_names = model["state_names"]
    model_params = get_params(model)

    # load all seqrecords into RAM
    seqrecords = load_seqrecords(seq_filename, include, exclude)
    assert seqrecords

    # load seqrecords2 if in diff mode
    seqrecords2 = load_seqrecords(seq2_filename, include, exclude)
    seqrecords2_dict = dict((defline_identifier(defline), seq)
                            for defline, seq in seqrecords2)
    assert (not simulate_func == simulate_diff) or seqrecords2_dict

    limit_slice_adj = correct_limit_slice(limit_slice)

    # init
    groupings = None
    if simulate_func == simulate_reference:
        measurement = "prob"
        cols_keep, colnames = match_names(keep_pattern, state_names)

        if total_stems:
            groupings, colnames = make_groupings(colnames)

        cells_keep = (limit_slice_adj, cols_keep)
    else:
        # simulate_mutations or simulate_diff
        measurement = "sum_relent_ref_mut"

        # with no cols specified, this is just the rows to keep:
        cells_keep = limit_slice_adj
        colnames = array(ALPHABET)

    # process
    driver = driver_class(args, colnames, measurement, write_seq,
                          start_default, limit_slice)
    with driver as driver:
        for seqrecord in chunk(seqrecords):
            defline, seq = seqrecord

            if seqrecords2_dict:
                seq2 = seqrecords2_dict[defline_identifier(defline)]

                # overload seqrecord with both seqs
                seqrecord = (defline, array([seq, seq2]))
            try:
                simulate_seq(simulate_func, driver, model_params, num_states,
                             seqrecord, cells_keep, groupings)
            except KeyError: # outside ALPHABET, such as N
                msg = "%s contains illegal characters; skipping" % defline
                print >>sys.stderr, msg

# XXX: this is a clunky way to do the driver options. Instead, why
# don't we do one option parse and repeat it with combined options if
# there is a driver. will need special handling for --help (and
# double-check conflict handling), but can eliminate driver-help
def parse_options(args):
    from ._optparse import OptionGroup, OptionParser

    usage = """%prog [OPTION...] MODEL SEQFILE OUTFILE
       %prog [OPTION...] --driver=DRIVER MODEL SEQFILE [DRIVEROPTIONS...]"""
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.disable_interspersed_args()

    with OptionGroup(parser, "Model") as group:
        group.add_option("-R", "--resource", action="store_true",
                         help="get MODEL from a resource that comes with the"
                         " sunflower distribution instead of a file")

    with OptionGroup(parser, "Sequence") as group:
        group.add_option("-i", "--include", action="append", default=[],
                         metavar="IDENTIFIER", help="run only on IDENTIFIER")
        group.add_option("", "--include-from", action="load", dest="include",
                         metavar="FILE",
                         help="run only on specified identifiers from FILE")
        group.add_option("-x", "--exclude", action="append", default=[],
                         metavar="IDENTIFIER", help="don't run on IDENTIFIER")
        group.add_option("", "--exclude-from", action="load", dest="exclude",
                         metavar="FILE",
                         help="don't run on specified identifiers from FILE")

    with OptionGroup(parser, "Comparison mode") as group:
        group.add_option("-m", "--mutate", action="store_const",
                         dest="simulator",
                         const=simulate_mutations, default=simulate_reference,
                         help="iteratively mutate input sequences")

        group.add_option("-2", "--diff", metavar="SEQFILE2",
                         help="compute difference between sequences in SEQFILE"
                         " and SEQFILE2")

    with OptionGroup(parser, "Output options") as group:
        group.add_option("-s", "--start", type=int,
                         metavar="POS", help="call the first position POS")
        group.add_option("-S", "--states", metavar="REGEX",
                         default=r"^(?!silent$)",
                         help="only write states matching REGEX")
        group.add_option("--separate-state-groups", action="store_true",
                         help="don't add together .f0, .f1, .r0, etc.")
        group.add_option("-r", "--range", type=slice, metavar="SLICE",
                         default=slice(None),
                         help="only write positions in SLICE "
                         "(ignores start values)")
        group.add_option("--no-sequence", action="store_true",
                         help="don't write seq attribute of output "
                         "(saves space)")

    with OptionGroup(parser, "Output driver") as group:
        group.add_option("-d", "--driver", default="DEFAULT",
                         help="set output format")

        group.add_option("--driver-help", action="store_true",
                         help="get driver-specific help")

    options, args = parser.parse_args(args)

    if options.driver_help:
        driver_class = driver_load(options.driver)
        driver_class.print_help()
        sys.exit(0)

    if not len(args) >= 2:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    driver_class = driver_load(options.driver)

    hmm_filename = args[0]
    seq_filename = args[1]

    if options.diff:
        # can't use --mutate and --diff together
        assert options.simulator == simulate_reference
        options.simulator = simulate_diff

    return simulator(hmm_filename, seq_filename, args[2:],
                     simulate_func=options.simulator,
                     driver_class=driver_class,
                     keep_pattern=options.states,
                     include=options.include, exclude=options.exclude,
                     resource=options.resource,
                     start_default=options.start,
                     limit_slice=options.range,
                     write_seq=(not options.no_sequence),
                     total_stems=(not options.separate_state_groups),
                     seq2_filename=options.diff)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

"""
split: DESCRIPTION

OUTTMPL like myfile.%s.h5
"""

__version__ = "$Revision: 413 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from contextlib import nested
from itertools import repeat
import sys

from tables import Group, NoSuchNodeError

from .hdf5 import h5open, walk_nodes
from .utils import progress

#parameters.NODE_MAX_SLOTS = 0

def split_file(inwalker, outfiles, measurements):
    for src_node in inwalker:
        name = src_node.name
        progress(name)

        assert (src_node.attrs.measurement == measurements).all()

        src_array = src_node.read()
        title = src_node.title
        byteorder = src_node.byteorder
        group_pathname = src_node._v_parent._v_pathname
        src_attrs = src_node.attrs

        for outfile_index, outfile in enumerate(outfiles):
            dst_array = src_array[..., outfile_index]

            dst_node = outfile.createArray(group_pathname, name, dst_array,
                                           title, byteorder,
                                           createparents=True)

            src_attrs._f_copy(dst_node)
            dst_node.attrs.measurement = measurements[outfile_index]

def split(infilename, out_template, inroot_pathname="/"):
    if inroot_pathname is None:
        inroot_pathname = "/"

    with h5open(infilename) as infile:
        inroot_node = infile.getNode(inroot_pathname)

        # using isinstance to avoid running afoul of natural namings
        # that mess this up
        if isinstance(inroot_node, Group):
            inwalker_factory = inroot_node._f_walkNodes
        else:
            inwalker_factory = repeat(inroot_node)

        first_array = inwalker_factory().next()

        measurements = first_array.attrs.measurement

        assert not isinstance(measurements, basestring)

        outfile_managers = (h5open(out_template % measurement, "a")
                            for measurement in measurements)

        with nested(*outfile_managers) as outfiles:
            split_file(inwalker_factory(), outfiles, measurements)

def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... INFILE OUTTMPL"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-n", "--node", metavar="NODE",
                      help="only produce output for children of NODE")

    options, args = parser.parse_args(args)

    if not len(args) == 2:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    return split(args[0], args[1], inroot_pathname=options.node)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division

"""
hdf5: HDF5/PyTables utility functions
"""

__version__ = "$Revision: 428 $"

from collections import defaultdict
from contextlib import closing
from hashlib import md5
import sys
from warnings import simplefilter

import tables

from ._utils import progress

SEP = "/" # XXX: this should be in PyTables
PREFIX = "_"
NYBBLES = 2

simplefilter("ignore", tables.NaturalNameWarning)

def balanced_path(name, nybbles=NYBBLES):
    hexdigest = md5(name).hexdigest()

    return SEP + PREFIX + hexdigest[:nybbles]

def balanced_path_full(name, *args, **kwargs):
    return SEP.join([balanced_path(name, *args, **kwargs), name])

# XXXopt: caching these here may be a memory leak
class H5GroupDefaultdict(defaultdict):
    def __init__(self, h5file, *args, **kwargs):
        self.h5file = h5file
        defaultdict.__init__(self, object, *args, **kwargs)

    def __missing__(self, key):
        res = self.h5file.createGroup("/", key.lstrip("/"))
        self[key] = res

        return res

class H5BalancedGroupCache(object):
    def __init__(self, h5file):
        self.h5file = h5file
        self.groups = H5GroupDefaultdict(h5file)

    def __getitem__(self, name):
        return self.groups[balanced_path(name)]

    def create_dataset(self, class_name, name, *args, **kwargs):
        where = self[name]._v_pathname

        method = getattr(self.h5file, "create" + class_name)
        return method(where, name, *args, **kwargs)

# XXX: upgrade PyTables, remove this
def h5open(filename, *args, **kwargs):
    return closing(tables.openFile(filename, *args, **kwargs))

def get_node(h5file, name):
    return h5file.getNode(balanced_path(name), name)

# slightly depth first for performance reasons
# assumes the layout is all groups on the first layer
def walk_nodes(h5file, classname="Array"):
    for group in h5file.root._v_groups.itervalues():
        for node in group._f_walkNodes(classname):
            yield node

def walk_nodes_progress(*args, **kwargs):
    inarrays = walk_nodes(*args, **kwargs)

    for inarray_index, inarray in enumerate(inarrays):
        progress("%6d: %s", inarray_index, inarray.name)

        yield inarray

def walk_included_nodes(h5file, names):
    if names:
        return (get_node(h5file, name) for name in names)
    else:
        return walk_nodes(h5file)

if __debug__:
    __pychecker__ = "no-special"



#!/usr/bin/env python
from __future__ import division, with_statement

"""
report: aggregate by sequence or position
"""

__version__ = "$Revision: 457 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import csv
from itertools import izip, repeat
import sys

from numpy import (NINF, PINF, array, errstate, finfo, histogram,
                   linspace, log1p, minimum, maximum, ndindex, newaxis,
                   rollaxis, square, sqrt as asqrt, zeros)
import tabdelim # for side-effect: dialect="unix-tab"

from ._utils import correct_range, enum, progress
from .hdf5 import h5open, walk_included_nodes, walk_nodes
from .factorizers import (AlignedOnlyAlignmentFactorizer,
                          ByNucleotideAlignmentFactorizer,
                          ByPosAlignmentFactorizer,
                          LABEL_ALIGNED_ONLY_ALIGNED, LABEL_ALIGNED_ONLY_OBS,
                          SubsOnlyAlignmentFactorizer, choose_factorizer,
                          fmt_float, calc_total_datacol)

FIELDNAMES = ["seqname", "factor", "datacol", "pos", "measurement",
              "sd", "min", "max", "n", "T", "T_d", "d_T"]

# layout of stats before transposition:
# stat: (mean, sd, min, max, n)
# datacol: state (reference mode) or nucleotide (mutate mode)

AXES_ENUM = enum(["seq", "stat", "factor", "pos", "datacol"])
AXIS_SEQ = AXES_ENUM["seq"]
AXIS_STAT = AXES_ENUM["stat"]
AXIS_FACTOR = AXES_ENUM["factor"]
AXIS_POS = AXES_ENUM["pos"]
AXIS_DATACOL = AXES_ENUM["datacol"]

STAT_OFFSET_ENUM = enum(["mean", "sd", "min", "max", "n"])
STAT_OFFSET_MEAN = STAT_OFFSET_ENUM["mean"]
STAT_OFFSET_N = STAT_OFFSET_ENUM["n"]

NUM_BINS_DEFAULT = 100

def distance_jc69(p):
    """calculate Jukes and Cantor 1969 distances"""
    return -0.75*log1p((-4/3)*p)

def calc_total_ranges(addend_array, ranges, methodname="sum"):
    totals = [getattr(addend_array[:, r], methodname)(1) for r in ranges]

    res = array(totals) # axes = (range, factor, datacol)
    return res.transpose(1, 0, 2) # axes = (factor, range, datacol)

def correct_ranges(ranges, start):
    return [correct_range(r, start) for r in ranges]

def calc_total(addend_array, ranges, total_datacol, methodname="sum"):
    if ranges:
        addend_array = calc_total_ranges(addend_array, ranges, methodname)
    if total_datacol:
        addend_array = calc_total_datacol(addend_array, methodname)

    return addend_array

class ArrayRecordList(list):
    """dummy list subclass to allow setting of attributes"""

class FactorizedArrayChecker(object):
    """
    checks that shape and start are consistent
    """
    def __init__(self, arrays):
        self.arrays = arrays

    def __iter__(self):
        shape = None

        for seqname, transcript_array, attrs in self.arrays:
            if shape is None:
                shape = transcript_array.shape
                self.shape = shape

                start = attrs.start
                self.start = start

                # ensure that there are no extra columns for various
                # measurements--unsupported
                assert len(shape) == 3
            else:
                assert shape == transcript_array.shape
                assert start == attrs.start

            yield seqname, transcript_array

    def wrap(self, arrayrecord):
        res = ArrayRecordList([arrayrecord])
        res.shape = self.shape
        res.start = self.start

        return res

def histogram_arrays(factorized_arrays, ranges, total_datacol, num_bins,
                     bin_range, measurement_min):
    assert measurement_min == 0.0
    """
    overload AXIS_STAT to return different bins
    """
    start = None

    # transcript_array: (factor, pos, datacol)
    for seqname, transcript_array in factorized_arrays:
        if start is None:
            # not determined until first array is read
            start = factorized_arrays.start
            len_factor, _, len_datacol = factorized_arrays.shape
            ranges = correct_ranges(ranges, start)
            len_ranges = len(ranges)

            if total_datacol:
                len_datacol_out = 1
            else:
                len_datacol_out = len_datacol

            # (factor, bin, datacol, range) or (factor, pos, datacol, stat)
            res_shape = (len_factor, len_ranges, len_datacol_out, num_bins)
            res = zeros(res_shape, dtype=int)

        # generates (0, 0, 0) .. (n, n, n)
        for indices in ndindex(len_factor, len_ranges, len_datacol):
            r = ranges[indices[1]]
            range_array = transcript_array[indices[0], r, indices[2]]

            # _ is a common Python convention for an unused variable
            range_histogram, _ = histogram(range_array, num_bins, bin_range)

            if total_datacol:
                indices = indices[:2] + (0,)

            res[indices] += range_histogram

    # (stat, factor, pos, datacol)
    return rollaxis(res, 3, 0), start

def aggregate_arrays(factorized_arrays, ranges, total_datacol, calc_d_T,
                     measurement_min):
    # XXXopt: if you don't want min, max, sd, many calcs can be avoided.
    # But for now this is the fastest part of the program anyway.

    sums = 0.0
    square_sums = 0.0
    divisors = 0
    mins = PINF
    maxs = NINF

    for seqname, transcript_array in factorized_arrays:
        # replace zeros (or other ignored number) with +inf
        transcript_array_is_zero = transcript_array <= measurement_min
        transcript_array_zeros_replaced = transcript_array.copy()
        transcript_array_zeros_replaced[transcript_array_is_zero] = PINF

        transcript_array[transcript_array_is_zero] = 0.0 # set ignored to zero

        # accumulate statistics from transcript_array
        sums += transcript_array
        square_sums += square(transcript_array)
        divisors += ~transcript_array_is_zero
        mins = minimum(mins, transcript_array_zeros_replaced)
        maxs = maximum(maxs, transcript_array)

    # must be after start is determined
    start = factorized_arrays.start
    ranges = correct_ranges(ranges, start)

    if calc_d_T:
        T_d = sums[newaxis, LABEL_ALIGNED_ONLY_OBS]
        T = sums[newaxis, LABEL_ALIGNED_ONLY_ALIGNED]

    sums = calc_total(sums, ranges, total_datacol)
    square_sums = calc_total(square_sums, ranges, total_datacol)
    divisors = calc_total(divisors, ranges, total_datacol)
    mins = calc_total(mins, ranges, total_datacol, "min")
    maxs = calc_total(maxs, ranges, total_datacol, "max")

    # calculating t or T by using the mean implies that t_d can
    # possibly be bigger than t for a given position, but it is
    # unlikely across a whole gene at the evolutionary distances the
    # whole endeavor will work. This sort of thing can happen in NG86
    # as well; see the example in Ch. 4 of Nei and Kumar for a TGT
    # codon
    if calc_d_T:
        T_d = calc_total(T_d, ranges, total_datacol)
        T = calc_total(T, ranges, total_datacol)
        if total_datacol:
            assert measurement_min == 0.0 # dividing by 3.0 won't work here
            T /= 3.0 # the mean of the three possible changes

    with errstate(invalid="ignore"):
        means = sums / divisors
        sds = calc_sd(sums, square_sums, divisors)

        if calc_d_T:
            p_T = T_d / T
            d_T = distance_jc69(p_T)

    if calc_d_T:
        res = [T, T_d, d_T]
    else:
        res = [means, sds, mins, maxs, divisors]

    # (stat, factor, pos, datacol)
    return array(res), start

def process_arrays(factorized_arrays, ranges, total_datacol, calc_d_T,
                   num_bins, bin_range, measurement_min):
    if num_bins:
        return histogram_arrays(factorized_arrays, ranges, total_datacol,
                                num_bins, bin_range, measurement_min)
    else:
        return aggregate_arrays(factorized_arrays, ranges, total_datacol,
                                calc_d_T, measurement_min)

def calc_stats_sequential(factorized_arrays, *args, **kwargs):
    seqnames = []
    starts = []
    stats_list = []

    for arrayrecord in factorized_arrays:
        seqname = arrayrecord[0]
        seqnames.append(seqname)

        wrapped_arrayrecord = factorized_arrays.wrap(arrayrecord)
        stats_seq, start = process_arrays(wrapped_arrayrecord, *args, **kwargs)
        stats_list.append(stats_seq)
        starts.append(start)

    return seqnames, starts, array(stats_list)

def calc_stats_aggregate(factorized_arrays, *args, **kwargs):
    seqnames = ["all"]
    stats_mean, start = process_arrays(factorized_arrays, *args, **kwargs)

    return seqnames, [start], array([stats_mean])

def calc_sd(sums, square_sums, divisors):
    # this will always give sample sd
    # maybe we should use population sd if no subset is specified?
    numerator = square_sums - (square(sums) / divisors)
    denominator = divisors - 1

    return asqrt(numerator / denominator)

def find_max(inarrays):
    res = NINF

    for inarray_index, inarray in enumerate(inarrays):
        if not inarray_index % 10:
            progress(" %5d", inarray_index)

        res = max(res, inarray.read().max())

    return res

def aggregate_filenames(h5filenames, factorizer, aggregate_func,
                        include_names, ranges, total_datacol, calc_d_T,
                        num_bins, hist_max, measurement_min):
    # XXX: make a generator to add arrays in additional filenames to inarrays
    assert len(h5filenames) == 1
    h5filename = h5filenames[0]

    with h5open(h5filename) as h5file:
        # XXX: start getting these from new file-wide attrs
        first_array = walk_included_nodes(h5file, include_names).next()
        first_attrs = first_array.attrs
        measurement = first_attrs.measurement
        if total_datacol:
            datacolnames = ["all"]
        else:
            datacolnames = list(first_attrs.colnames)

        if num_bins:
            # smallest magnitude usable number (to exclude zeros)
            tiny = finfo(first_array.read().dtype).tiny

            if hist_max is None:
                hist_max = find_max(walk_nodes(h5file))

            bin_range = (tiny, hist_max)
        else:
            bin_range = None

        inarrays = walk_included_nodes(h5file, include_names)
        factorized_arrays = factorizer.factor_arrays(inarrays)
        checked_factorized_arrays = FactorizedArrayChecker(factorized_arrays)

        # XXXopt: add a thread to start reading in inarrays ahead of time to
        #         ease I/O bottleneck
        seqnames, starts, stats = \
            aggregate_func(checked_factorized_arrays, ranges, total_datacol,
                           calc_d_T, num_bins, bin_range, measurement_min)

    return stats, measurement, datacolnames, seqnames, starts, bin_range

def write_header(writer, fieldnames, measurement):
    row = fieldnames[:]

    try:
        row[row.index("measurement")] = measurement
    except ValueError:
        pass

    try:
        if measurement == "prob":
            row[row.index("datacol")] = "state"
        else:
            row[row.index("datacol")] = "mutation"
    except ValueError:
        pass # no datacol column

    writer.writerow(row)

def make_fieldnames(stats, all_columns, calc_d_T, num_bins):
    fieldnames = FIELDNAMES[:]

    if not all_columns:
        shape = stats.shape

        if shape[AXIS_SEQ] == 1:
            fieldnames.remove("seqname")

        if shape[AXIS_FACTOR] == 1:
            fieldnames.remove("factor")

        if shape[AXIS_DATACOL] == 1:
            fieldnames.remove("datacol")

        if calc_d_T:
            fieldnames.remove("measurement")
        else:
            fieldnames.remove("T")
            fieldnames.remove("T_d")
            fieldnames.remove("d_T")

        if calc_d_T or num_bins or (stats[:, STAT_OFFSET_N] <= 1.0).all():
            fieldnames.remove("sd")
            fieldnames.remove("min")
            fieldnames.remove("max")

            if not num_bins:
                fieldnames.remove("n")

    return fieldnames

def slice_value2str(num):
    if num is None:
        return ""

    return str(num)

def range2str(r):
    if isinstance(r, slice):
        vals = [r.start, r.stop]
        if r.step is not None:
            vals.append(r.step)

        return ":".join(map(slice_value2str, vals))
    else:
        return str(r)

# XXX: too many arguments
def report(h5filenames, factorizer, aggregate_func, start_override=None,
           include_names=None, ranges=[], all_columns=False,
           total_datacol=False, calc_d_T=False, num_bins=None, hist_max=None,
           measurement_min=0.0):
    assert not (calc_d_T and num_bins)

    stats, measurement, datacolnames, seqnames, starts, bin_range = \
        aggregate_filenames(h5filenames, factorizer, aggregate_func,
                            include_names, ranges, total_datacol, calc_d_T,
                            num_bins, hist_max, measurement_min)

    if ranges:
        range_strs = [range2str(r) for r in ranges]
        start_override = True

    if num_bins:
        bin_edges = linspace(bin_range[0], bin_range[1], num_bins+1)
        bin_starts = bin_edges[:-1]

    if start_override:
        starts = repeat(start_override)

    fieldnames = make_fieldnames(stats, all_columns, calc_d_T, num_bins)
    fieldname_set = frozenset(fieldnames)

    writer = csv.writer(sys.stdout, dialect="unix-tab")
    write_header(writer, fieldnames, measurement)

    stats_transposed = stats.transpose(AXIS_SEQ, AXIS_FACTOR, AXIS_DATACOL,
                                       AXIS_POS, AXIS_STAT)

    seq_zipper = izip(seqnames, starts, stats_transposed)
    for seqname, start, stats_seq in seq_zipper:
        # arrays should be homogeneous from this point down

        # XXX: spaghetti code ahoy!
        # instead, make a recordarray and adjust in memory
        for factor_label, stats_factor in izip(factorizer, stats_seq):
            # izip() won't produce overlap
            for datacolname, stats_datacol in izip(datacolnames, stats_factor):
                for pos_raw, stats_pos in enumerate(stats_datacol):
                    if ranges:
                        pos = range_strs[pos_raw]
                    else:
                        pos = str(pos_raw + start)

                    if not (calc_d_T or num_bins):
                        pos_mean, pos_sd, pos_min, pos_max, pos_n = stats_pos

                        if pos_n == 0.0:
                            continue

                    outrow = []
                    if "seqname" in fieldname_set:
                        outrow.append(seqname)
                    if "factor" in fieldname_set:
                        outrow.append(factor_label)
                    if "datacol" in fieldname_set:
                        outrow.append(datacolname)

                    outrow.append(pos)

                    if calc_d_T:
                        outrow.extend(map(fmt_float, stats_pos))
                    elif num_bins:
                        # stats_pos is overloaded for bins
                        for bin_index, bin in enumerate(stats_pos):
                            outrow_bin = outrow[:]
                            outrow_bin.append(bin_starts[bin_index])
                            outrow_bin.append(bin)

                            writer.writerow(outrow_bin)

                        continue # don't write outrow again
                    else:
                        outrow.append(fmt_float(pos_mean))

                        if "sd" in fieldname_set:
                            cols = map(fmt_float, [pos_sd, pos_min, pos_max])
                            outrow.extend(cols)
                        if "n" in fieldname_set:
                            outrow.append("%d" % pos_n)

                    writer.writerow(outrow)

def parse_options(args):
    from ._optparse import OptionGroup, OptionParser

    usage = "%prog [OPTION]... FILE"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)

    with OptionGroup(parser, "Output control") as group:
        # XXX: eliminate this after another release
        group.add_option("-p", "--start", type=int,
                         metavar="POS",
                         help="[DEPRECATED] call the first position POS")
        group.add_option("-A", "--aggregate", action="store_const",
                         dest="aggregate_func",
                         default=calc_stats_sequential,
                         const=calc_stats_aggregate,
                         help="average results from different sequences "
                         "(sum in histogram mode)")
        group.add_option("-T", "--total", action="store_true",
                         help="total mutation or state")
        group.add_option("--no-total", action="store_false", dest="total",
                         help="report mutation or state individually")
        group.add_option("-C", "--all-columns", action="store_true",
                         help="include all columns in output, even if they are"
                         " uninformative")
        # XXX: should be an optional default value (100)
        group.add_option("-H", "--hist-bins", type=int, metavar="BINS",
                         help="output histograms of BINS bins")

    # all of these options can be included multiple times
    with OptionGroup(parser, "Specify a subset") as group:
        group.add_option("-i", "--include", action="append", dest="include",
                         metavar="ARRAY", help="report only on ARRAY")
        group.add_option("--include-from", action="load", dest="include",
                         metavar="FILE",
                         help="report only on arrays specified in FILE")

        # this will get you -50..49 == 100 positions
        group.add_option("-r", "--range", type=slice, default=[],
                         action="append", metavar="RANGE",
                         help='report only on average results for positions '
                         'RANGE in the form of "-50:50"')

        group.add_option("--hist-max", type=float, metavar="LIMIT",
                         help="calculate histogram bins in the range 0:LIMIT")

        group.add_option("--measurement-min", type=float, metavar="MIN",
                         help="only include data where measurement > MIN",
                         default=0.0)

    # XXX: UI probably a bit confusing here
    with OptionGroup(parser, "Stratification factor control") as group:
        group.add_option("-t", "--tab-delim", metavar="TABFILE",
                         help="stratify on identifiers in TABFILE")
        group.add_option("-a", "--alignment", metavar="FASTAFILE",
                         help="stratify on align slices in FASTAFILE")
        group.add_option("-d", "--distances", action="store_true",
                         help="calculate evolutionary distances")
        group.add_option("-P", "--by-position", action="store_const",
                         dest="alignment_factorizer",
                         const=ByPosAlignmentFactorizer,
                         default=ByNucleotideAlignmentFactorizer,
                         help="stratify by position; disregard nucleotide")
        group.add_option("-s", "--only-substitutions", action="store_const",
                         dest="alignment_factorizer",
                         const=SubsOnlyAlignmentFactorizer,
                         help="only include substitution positions")

    options, args = parser.parse_args(args)

    if len(args) != 1:
        parser.print_usage()
        sys.exit(1)

    if options.start and options.range:
        # XXX: allow this: if you use both, then correct_ranges in advance
        parser.error("can't use --start and --range together")

    if options.distances and options.hist_bins:
        parser.error("can't use --distances and --histogram together")

    if options.distances and not options.alignment:
        parser.error("must use --alignment when using --distances")

    if options.hist_max and not options.hist_bins:
        options.hist_bins = NUM_BINS_DEFAULT

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    calc_d_T = options.distances
    if calc_d_T:
        assert options.alignment

        # XXX: should eliminate the default in options, set it elsewhere
        # this means no specific option was set
        assert options.alignment_factorizer == ByNucleotideAlignmentFactorizer

        # override that
        options.alignment_factorizer = AlignedOnlyAlignmentFactorizer

    factorizer = choose_factorizer(options)
    include_names = options.include
    start_override = options.start
    aggregate_func = options.aggregate_func
    all_columns = options.all_columns
    ranges = options.range
    num_bins = options.hist_bins
    hist_max = options.hist_max
    total_datacol = options.total
    measurement_min = options.measurement_min

    # many other options are used in factorizers.py

    if num_bins and not ranges:
        ranges = [slice(None)]

    if total_datacol is None and ranges:
        total_datacol = True

    return report(args, factorizer, aggregate_func, start_override,
                  include_names, ranges, all_columns, total_datacol, calc_d_T,
                  num_bins, hist_max, measurement_min)

if __name__ == "__main__":
    sys.exit(main())

if __debug__:
    __pychecker__ = "unusednames=tabdelim"



#!/usr/bin/env python
from __future__ import division

__version__ = "$Revision: 366 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from itertools import izip
import re
import sys
from warnings import filterwarnings

from MySQLdb import IntegrityError
from dboptions import DbOptionParser, connect
import docsql
from numpy import empty, uint8

from . import Driver, DriverHelpFormatter, defline_identifier

QUANTIZATION_SCALE = 255.0

class Insert_tf(docsql.Insert):
    """
    INSERT INTO tf SET name=%s
    """

class Select_tf_id(docsql.SelectOneCell):
    """
    SELECT tf_id FROM tf WHERE name=%s
    """

class Insert_seq_region(docsql.Insert):
    """
    INSERT INTO seq_region SET name=%s
    """

class Select_seq_region_id(docsql.SelectOneCell):
    """
    SELECT seq_region_id FROM seq_region WHERE name=%s
    """

class Insert_prob(docsql.Insert):
    """
    INSERT DELAYED INTO prob (tf_id, seq_region_id, seq_region_start,
                              seq_region_end, probs)
    VALUES (%s, %s, %s, %s, %s)
    """

def get_rowid(insert, select, name, connection):
    try:
        cursor = insert(name, connection=connection)
        return cursor.lastrowid
    except IntegrityError:
        return select(name, connection=connection)

class RowAppender(object):
    """
    converts 0-based to offset 1-based and appends
    """

    def __init__(self, sequence, tf_id, seq_region_id, start):
        self.sequence = sequence
        self.tf_id = tf_id
        self.seq_region_id = seq_region_id
        self.start = start

    def __call__(self, start, end, values):
        self.sequence.append([self.tf_id,
                              self.seq_region_id,
                              start + self.start,
                              end + self.start - 1,
                              values])

# 19 bytes or greater
# XXX: there's a pathological case where you have something that looks like
# [19A, 1B, 19A]
# you would be better off representing it as
# AAAAAAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAAAA
# because the 1B uses so much space
#
# this would add extra code complexity to deal with properly; probably
# not worth it
re_charrun = re.compile(r"(.)\1{17,}")
class SQLDriver(Driver):
    def __init__(self, command_args, *args, **kwargs):
        options, command_args = self.parse_options(command_args)
        self.dboptions = options.dboptions

        Driver.__init__(self, [], *args, **kwargs)

    @staticmethod
    def get_option_parser():
        usage = ("%prog --driver=DRIVER [OPTION...] MODEL SEQFILE "
                 "[DRIVEROPTION...]")
        return DbOptionParser(usage=usage, formatter=DriverHelpFormatter())

    def parse_options(self, args):
        parser = self.get_option_parser()
        options, args = parser.parse_args(args)

        if len(args) != 0:
            self.print_usage()
            sys.exit(1)

        return options, args

    def __enter__(self):
        connection = connect(self.dboptions)
        self.connection = connection
        self.tf_ids = [get_rowid(Insert_tf, Select_tf_id, colname, connection)
                       for colname in self.colnames]

        return self

    def __exit__(self, *exc_info):
        self.connection.close()

    # XXXopt:
    # reimplement run-length encoding from
    # http://projects.scipy.org/pipermail/numpy-discussion/2007-October/029378.html
    def __call__(self, arr, (defline, seq)):
        name = defline_identifier(defline)
        connection = self.connection

        seq_region_id = get_rowid(Insert_seq_region, Select_seq_region_id,
                                  name, connection=connection)

        start = self.defline_start(defline)
        assert start >= 1

        # uses floor to quantize
        quantized_array = (arr * QUANTIZATION_SCALE).astype(uint8)

        rows = []
        for tf_id, tf_vec in izip(self.tf_ids, quantized_array.T):
            values = tf_vec.tostring()
            appender = RowAppender(rows, tf_id, seq_region_id, start)

            charrun_start = None
            charrun_end = 0

            for m_charrun in re_charrun.finditer(values):
                charrun_start = m_charrun.start()

                # append previous unencoded chunk
                if charrun_start != charrun_end:
                    values_chunk = values[charrun_end:charrun_start]
                    appender(charrun_end, charrun_start, values_chunk)

                charrun_end = m_charrun.end()

                # append encoded chunk
                appender(charrun_start, charrun_end, m_charrun.group(1))

            values_end = len(values)

            # append last unencoded chunk
            if charrun_start != values_end:
                values_chunk = values[charrun_end:]
                appender(charrun_end, values_end, values_chunk)

        Insert_prob(many=rows, connection=connection)

def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division

"""
_optparse: add an optparse action to load an option from a file, to
           add to dicts, and a type for slices
"""

__version__ = "$Revision: 275 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from collections import defaultdict
from optparse import (Option as _Option, OptionError,
                      OptionGroup as _OptionGroup,
                      OptionParser as _OptionParser, OptionValueError)
import sys

def _constant(constant):
    def _constant_inner():
        return constant

    return _constant_inner

# adapted from http://docs.python.org/lib/optparse-callback-example-6.html
def optional_arg_callback(option, opt_str, value, parser):
    assert value is None

    done = 0
    value = []
    rargs = parser.rargs

    # Stop if we hit an arg like "--foo", "-a", "-fx", "--file=f",
    # etc.  Note that this also stops on "-3" or "-3.0", so if
    # your option takes numeric values, you will need to handle
    # this.

    # XXX: extract block into a function to reduce spaghetti
    if rargs and rargs[0] and not rargs[0].startswith("-"):
        try:
            # XXX: allow non-float types
            value = float(rargs[0])
            del rargs[0]
        except ValueError:
            value = option.unspecified
    else:
        value = option.unspecified

    setattr(parser.values, option.dest, value)

def check_slice(option, opt, value):
    """
    returns an int or slice
    """
    try:
        args = [int(x) if x else None for x in value.split(":")]

        if len(args) == 1:
            return args[0]
        else:
            return slice(*args)

    except ValueError:
        message = "option %s: invalid slice value: %r" % (opt, value)
        raise OptionValueError(message)

class Option(_Option):
    ATTRS = _Option.ATTRS + ["unspecified"]
    ACTIONS = _Option.ACTIONS + ("load", "update", "update_const",
                                 "update_true", "update_false")
    STORE_ACTIONS = _Option.STORE_ACTIONS + ("load", "update", "update_const",
                                             "update_true", "update_false",
                                             "callback")
    TYPED_ACTIONS = _Option.TYPED_ACTIONS + ("load", "update")
    ALWAYS_TYPED_ACTIONS = _Option.ALWAYS_TYPED_ACTIONS + ("load", "update")
    NARGS_SPECIFIED_ACTIONS = defaultdict(_constant(None), update=2)
    CONST_ACTIONS = set(["store_const", "update_const"])

    # handle type conversion in take_action() not in process()
    TYPE_HANDLING_ACTIONS = set(["load", "update"])

    TYPES = _Option.TYPES + ("slice",)
    TYPE_CHECKER = _Option.TYPE_CHECKER.copy()
    TYPE_CHECKER["slice"] = check_slice

    def _check_const(self):
        if self.action not in self.CONST_ACTIONS and self.const is not None:
            msg = "'const' must not be supplied for action %r" % self.action
            raise OptionError(msg, self)

    def _check_nargs(self):
        nargs_limit = self.NARGS_SPECIFIED_ACTIONS[self.action]
        if nargs_limit is None:
            _Option._check_nargs(self)
        elif self.nargs is None:
            self.nargs = nargs_limit
        elif self.nargs != nargs_limit:
            msg = "'nargs' must be %d for action %r" % (nargs_limit,
                                                        self.action)
            raise OptionError(msg, self)

    CHECK_METHODS = [_Option._check_action,
                     _Option._check_type,
                     _Option._check_choice,
                     _Option._check_dest,
                     _check_const,
                     _check_nargs,
                     _Option._check_callback]

    # defer type conversion
    def process(self, opt, value, values, parser):
        if self.action in self.TYPE_HANDLING_ACTIONS:
            return self.take_action(self.action, self.dest, opt,value, values,
                                    parser)
        else:
            return _Option.process(self, opt, value, values, parser)

    def _update(self, dest, key, dict_value, values):
        values.ensure_value(dest, {})[key] = dict_value

    def take_action(self, action, dest, opt, value, values, parser):
        if action == "load":
            infile = open(value)
            add_values = (self.convert_value(opt, line.rstrip())
                          for line in infile)
            values.ensure_value(dest, []).extend(add_values)
        elif action == "update":
            key = value[0]
            dict_value = self.convert_value(opt, [value[1]])[0]
            self._update(dest, key, dict_value, values)
        elif action == "update_const":
            self._update(dest, key, self.const, values)
        elif action == "update_true":
            self._update(dest, key, True, values)
        elif action == "update_false":
            self._update(dest, key, False, values)
        else:
            _Option.take_action(self, action, dest, opt, value, values, parser)

# XXX: add list of defaults in help
class OptionParser(_OptionParser):
    def __init__(self, usage=None, option_list=None,
                 option_class=Option, *args, **kwargs):
        _OptionParser.__init__(self, usage, option_list, option_class, *args,
                               **kwargs)

class OptionGroup(_OptionGroup):
    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.parser.add_option_group(self)

def main(args):
    pass

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
from __future__ import division

__version__ = "$Revision: 428 $"

# Copyright 2007, 2009 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import sys

from tables import openFile

from . import FileDriver, defline_identifier
from .._utils import DESC
from ..hdf5 import H5BalancedGroupCache

class HDF5Driver(FileDriver):
    def __enter__(self):
        outfile = openFile(self.filename, "w", DESC)
        self.file = outfile
        self.groups = H5BalancedGroupCache(outfile)

        return self

    def __exit__(self, *exc_info):
        self.file.close()

    def __call__(self, arr, (defline, seq)):
        name = defline_identifier(defline)
        leaf = self.groups.create_dataset("Array", name, arr, defline)

        attrs = leaf.attrs
        attrs.measurement = self.measurement
        attrs.colnames = self.colnames
        attrs.run_info = DESC

        if self.write_seq:
            attrs.seq = seq

        attrs.start = self.defline_start(defline)

def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division

"""
composition: take composition into account

requires output from fastacomposition, from the exonerate package
"""

__version__ = "$Revision: 425 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from collections import defaultdict
import sys

from numpy import array, float64, log as alog

from ._utils import DESC
from .hmm import ALPHABET, adjust, load, save

PSEUDOCOUNT = 1

def add_pseudocount(n=0):
    return n + PSEUDOCOUNT

def cols2counts(cols):
    # cols = ['A', '578', 'C', '476', 'G', '485']
    # why only ACG? I'm demonstrating a corner case--it will still
    # work and will create a pseudocount for T

    labels = cols[::2]

    count_generator = (add_pseudocount(int(col)) for col in cols[1::2])

    # counts["X"] where X was previously undefined returns the pseudocount
    counts = defaultdict(add_pseudocount, zip(labels, count_generator))

    return array([counts[letter] for letter in ALPHABET])

def load_fastacomposition(filename):
    # examples:
    # line = "test.fna A 578 C 476 G 485"
    line = iter(open(filename)).next()

    # cols = ['test.fna', 'A', '578', 'C', '476', 'G', '485']
    return line.rstrip().split()[1:]

def calc_logprobs_emission(counts):
    counts_float = array(counts, dtype=float64)

    return alog(counts_float / counts.sum())

def cols2logprobs_emission(cols):
    return calc_logprobs_emission(cols2counts(cols))

def recompose(composition_filename, infilename, outfilename, resource=False):
    cols = load_fastacomposition(composition_filename)
    logprobs_emission = cols2logprobs_emission(cols)

    model = load(infilename, resource)
    adjust(model, logprobs_emission)
    model["desc"] = "; ".join([model["desc"], DESC])

    save(outfilename, model)

def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... FASTACOMPOSITION IN-SFL OUT-SFL"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-R", "--resource", action="store_true",
                      help="get IN-SFL from a resource that comes with the "
                      "Sunflower distribution instead of a file")

    options, args = parser.parse_args(args)

    if not len(args) == 3:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    # XXX: allow more than two in/out pairs
    options, args = parse_options(args)

    return recompose(resource=options.resource, *args)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

"""
pwm2sfl: make pickled HMM from JASPAR-like source

default SPECIES is "Homo sapiens"
matrices that have no species specified (e.g. TBP) are always included

Prints out some things just for debugging purposes really. The output
is in whatever is specified by --output, jaspar.sfl by default.
"""

__version__ = "$Revision: 457 $"

# Copyright 2006-2009 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from math import log
import os
import sys

from numpy import NINF, array, float64, log as alog
from path import path

from ._utils import DESC
from .hmm import (ALPHABET, TF_START_SUFFIXES, LEN_ALPHABET, OFFSET_SILENT,
                  OFFSET_UNBOUND, SPECIAL_STATE_NAMES, fill_array, optimize,
                  save)

GC_SET = set("GC")

PWM_LIST_FILENAME = "matrix_list.txt"
PWM_EXT = "pfm"

# suffixes that mark the last begininng of a TF
TF_START_FINAL_SUFFIXES = set(["0", "r0"])

DEFAULT_GC_FRAC = (1.0 / LEN_ALPHABET) * 2
PSEUDOCOUNT = 1.0

DEFAULT_PROB_SILENT2UNBOUND = 0.9

# default mean distance is 2 helix-turns
DEFAULT_COOP_DISTANCE = 22

# default co-operativity strength boost is 25%
DEFAULT_COOP_STRENGTH = 0.25

OUTPUT_FILENAME = "jaspar.sfl"

class AppendableSet(set):
    def append(self, item):
        self.add(item)

class PwmInfo(object):
    def __init__(self, line):
        cols = line.rstrip().split("\t")

        (self.id, self.information_content, self.name, self.kind, other) = cols

        fields = other.lstrip("; ").split(" ; ")
        attribute_pairs = (field.split(" ") for field in fields)
        attribute_pairs_fmt = ((pair[0], self._fmt_attribute_value(pair[1:]))
                               for pair in attribute_pairs)

        self.__dict__.update(dict(attribute_pairs_fmt))

    @staticmethod
    def _fmt_attribute_value(value):
        return " ".join(value).strip('"')

def read_pwm(iterable):
    # each line is one of [A, C, G, T]
    for line in iterable:
        yield map(float, line.rstrip().split())

def make_pwm_filename(dirpath, id):
    return os.extsep.join([dirpath / id, PWM_EXT])

def load_pwm(dirpath, id):
    filename = os.extsep.join([dirpath / id, PWM_EXT])

    with open(filename) as infile:
        freq_pwm_list = list(read_pwm(infile))

    return array(freq_pwm_list)

def load_info(dirpath):
    filepath = dirpath / PWM_LIST_FILENAME

    with open(filepath) as infofile:
        for line in infofile:
            yield PwmInfo(line)

def make_state_names(pwms, motif_infos):
    state_names = SPECIAL_STATE_NAMES[:]

    # offset at which each tf starts
    # e.g. TF0.f0, TF0.r0, TF1.f0...
    tf_start_offsets = []

    # ...and ends
    tf_end_offsets = []

    for (pwm, (name, dir_tag)) in zip(pwms, motif_infos):
        tf_length = pwm.shape[1]
        tf_start_offsets.append(len(state_names))

        state_names.extend(".".join([name, "%s%d" % (dir_tag, index)])
                           for index in xrange(tf_length))

        tf_end_offsets.append(len(state_names)-1)

    return state_names, tf_start_offsets, tf_end_offsets

def attr_in_container(item, attr, container):
    """returns True if:
    * item does not have attr; or
    * attr is in container"""
    return not hasattr(item, attr) or getattr(item, attr) in container

def load_info_limited(dirpath, species_set, sysgroup_set):
    if not (species_set or sysgroup_set):
        for motif in load_info(dirpath):
            yield motif
    else:
        for motif in load_info(dirpath):
            if (attr_in_container(motif, "species", species_set)
                    or attr_in_container(motif, "sysgroup", sysgroup_set)):
                yield motif

def load_pwms(dirname, species_set, sysgroup_set, revcom):
    pwms = []
    motif_infos = []

    dirpath = path(dirname)

    isatty = sys.stderr.isatty()

    for motif in load_info_limited(dirpath, species_set, sysgroup_set):
        row = [len(pwms), motif.id, motif.name, motif.kind]
        if isatty:
            print >>sys.stderr, "\t".join(map(str, row))

        pwm = load_pwm(dirpath, motif.id)
        motif_name = motif.name

        pwms.append(pwm)

        if revcom:
            # this only works for ACGT
            assert ALPHABET == "ACGT"

            # reverse both the rows and the cols
            pwm_rev = pwm[::-1, ::-1]
            pwms.append(pwm_rev)

            # motif_info = (name, dir_tag)
            motif_infos.append((motif_name, "f"))
            motif_infos.append((motif_name, "r"))
        else:
            # if revcom is off, don't prepend f/r to sub-TF ID
            motif_infos.append((motif_name, ""))

    state_names, tf_start_offsets, tf_end_offsets = \
        make_state_names(pwms, motif_infos)

    return pwms, state_names, tf_start_offsets, tf_end_offsets

def add_cooperativity(model, coop_distance, coop_strength, probs_prior,
                      logprobs_e_background, tf_start_offsets, tf_end_offsets):
    ## make state names
    state_names = model["state_names"]
    coop_offsets = [] # offsets of cooperative states
    tf_coop_offsets = [] # one-to-one mapping with tf_start_offsets

    # copy so you don't iterate over the new stuff
    for state_name in state_names[:]:
        state_name_partition = state_name.rpartition(".")
        suffix = state_name_partition[2]

        if suffix in TF_START_SUFFIXES:
            num_states = len(state_names)
            tf_coop_offsets.append(num_states)

            if suffix in TF_START_FINAL_SUFFIXES:
                coop_offsets.append(num_states)
                state_names.append("%s.c" % state_name_partition[0])

    ## make matrices
    num_states = len(state_names)
    logprobs_a_old = model["logprobs_transition"]
    logprobs_e_old = model["logprobs_emission"]

    # RAM usage doubles here. logprobs_a needs to be a more efficient
    # data structure since it is sparse

    # XXX: move outer pwm2sfl() stuff into an inner function and then
    #      use resize() here
    logprobs_a = fill_array(NINF, (num_states, num_states))
    logprobs_e = fill_array(NINF, (num_states, LEN_ALPHABET))

    logprobs_a_old_shape = logprobs_a_old.shape
    logprobs_e_old_shape = logprobs_e_old.shape

    logprobs_a[:logprobs_a_old_shape[0], :logprobs_a_old_shape[1]] = \
        logprobs_a_old
    logprobs_e[:logprobs_e_old_shape[0], :logprobs_e_old_shape[1]] = \
        logprobs_e_old

    # calculate the probability of remaining in the coop state from the
    # mean distance parameter
    prob_coop2notcoop = 1.0 / coop_distance
    prob_coop2coop = 1.0 - prob_coop2notcoop

    prob_prior_unbound = probs_prior["unbound"]

    silent_strength = 1.0 - coop_strength
    prob_coop2silent = prob_coop2notcoop * prob_prior_unbound * silent_strength

    prob_coop2alltfs = 1 - prob_coop2coop - prob_coop2silent
    prob_silent2alltfs = 1 - prob_prior_unbound

    # take the logs
    logprob_coop2coop = log(prob_coop2coop)
    logprob_coop2silent = log(prob_coop2silent)

    ## transition inbound
    # don't vectorize until format change
    for coop_offset, tf_end_offset in zip(tf_coop_offsets, tf_end_offsets):
        # override previous setting:
        logprobs_a[tf_end_offset, OFFSET_SILENT] = NINF
        logprobs_a[tf_end_offset, coop_offset] = 0

    ## transition outbound
    for coop_offset in coop_offsets:
        logprobs_a[coop_offset, coop_offset] = logprob_coop2coop
        logprobs_a[coop_offset, OFFSET_SILENT] = logprob_coop2silent

        for tf_start_offset in tf_start_offsets:
            prob_prior_tf = probs_prior[state_names[tf_start_offset]]
            prob_prior_tf_scaled = prob_prior_tf * prob_coop2alltfs
            prob_coop2tf = prob_prior_tf_scaled / prob_silent2alltfs

            logprobs_a[coop_offset, tf_start_offset] = log(prob_coop2tf)

    ## emission
    logprobs_e[coop_offsets] = logprobs_e_background

    model.update(state_names=state_names,
                 logprobs_transition=logprobs_a,
                 logprobs_emission=logprobs_e)

def get_priors(prob_silent2unbound, state_names, tf_start_offsets, num_tfs,
               filename, revcom):
    probs_prior = {}

    if filename:
        ## read the file - must be in the form TF prior
        prior_init = {}
        with open(filename) as infile:
            for line in infile:
                tokens = line.split()
                if len(tokens) == 2:
                    prior_init[tokens[0]] = float(tokens[1])

        ## check that priors sum to 1 - if not, scale them
        total = sum(prior_init.values())
        if total != 1.0:
            # XXX: vectorize
            for key in prior_init.keys():
                prior_init[key] = prior_init[key] / total

        ## map the unbound prior
        probs_prior["unbound"] = prior_init["unbound"]

        ## map the tf priors
        for tf_start_offset in tf_start_offsets:
            tf_initial_state_name = state_names[tf_start_offset]
            if revcom:
                ## form .f0 or .r0; divide probability equally between
                ## forward and reverse
                prior_init_all = prior_init[tf_initial_state_name[:-3]]
                probs_prior[tf_initial_state_name] = prior_init_all / 2
            else:
                ## form .0
                prior_init_all = prior_init[tf_initial_state_name[:-2]]
                probs_prior[tf_initial_state_name] = prior_init_all

    else:
        ## only the unbound prior was specified
        probs_prior["unbound"] = prob_silent2unbound

        ## equally divide the rest of the prior among the transcription factors
        for tf_start_offset in tf_start_offsets:
            tf_state_name = state_names[tf_start_offset]
            probs_prior[tf_state_name] = (1 - prob_silent2unbound) / num_tfs

    return probs_prior

# XXX: break up into subfunctions
def pwm2sfl(dirname, prob_silent2unbound, priors,
            coop, coop_distance, coop_strength,
            revcom, species_set, sysgroup_set,
            gc_frac, output_filename=None):

    pwms, state_names, tf_start_offsets, tf_end_offsets = \
        load_pwms(dirname, species_set, sysgroup_set, revcom)

    num_states = len(state_names)
    num_tfs = len(tf_start_offsets)

    ## prior probabilities:
    ## either prob_silent2unbound + state_names + tf_start_offsets
    ## or priors (file)
    probs_prior = get_priors(prob_silent2unbound, state_names,
                             tf_start_offsets, num_tfs, priors, revcom)

    logprobs_a = fill_array(NINF, (num_states, num_states))
    logprobs_e = fill_array(NINF, (num_states, LEN_ALPHABET))

    ## silent
    logprobs_a_silent = logprobs_a[OFFSET_SILENT]
    logprobs_a_silent[OFFSET_SILENT] = NINF

    ## silent to unbound
    logprobs_a_silent[OFFSET_UNBOUND] = log(probs_prior["unbound"])

    ## silent to bound
    for tf_start_offset in tf_start_offsets:
        tf_state_name = state_names[tf_start_offset]
        logprobs_a_silent[tf_start_offset] = log(probs_prior[tf_state_name])

    # background distribution
    # XXX: note that this does not allow G != C or A != T
    # this only works for ACGT
    assert ALPHABET == "ACGT"

    # XXX: vectorize
    logprobs_e_background = [0.0] * LEN_ALPHABET
    for i in xrange(LEN_ALPHABET):
        base = ALPHABET[i]
        if base in GC_SET:
            logprobs_e_background[i] = log(gc_frac / 2.0)
        else:
            logprobs_e_background[i] = log((1 - gc_frac) / 2.0)

    ## unbound
    # remember: logprob = 0 means prob = 1
    logprobs_a[OFFSET_UNBOUND, OFFSET_SILENT] = 0
    logprobs_e[OFFSET_UNBOUND] = logprobs_e_background

    ## bound: transition
    # don't vectorize since the a format will change soon
    for offset in xrange(tf_start_offsets[0], num_states):
        if offset in tf_end_offsets:
            logprobs_a[offset, OFFSET_SILENT] = 0
        else:
            logprobs_a[offset, offset+1] = 0

    ## bound: emission
    # XXX: vectorize
    for pwm, offset in zip(pwms, tf_start_offsets):
        for index, col in enumerate(pwm.transpose()):
            # add as int32, then convert to float64:
            pseudocol = array(col + PSEUDOCOUNT, dtype=float64)
            total_counts = pseudocol.sum()

            logprobs_e[offset+index] = alog(pseudocol / total_counts)

    ## construct the model
    model = dict(alphabet=ALPHABET,
                 desc=DESC,
                 state_names=state_names,
                 logprobs_transition=logprobs_a,
                 logprobs_emission=logprobs_e)

    ## cooperativity
    if coop:
        add_cooperativity(model, coop_distance, coop_strength,
                          probs_prior, logprobs_e_background,
                          tf_start_offsets, tf_end_offsets)

    ## optimize and save the model
    optimize(model)
    save(output_filename, model)

def add_optgroup_input_limiters(parser):
    from ._optparse import OptionGroup

    with OptionGroup(parser, "Input limiters") as group:
        group.add_option("-s", "--species", action="append",
                         default=AppendableSet(), metavar="SPECIES",
                         help="include matrices marked from SPECIES")
        group.add_option("-S", "--sysgroup", action="append",
                         default=AppendableSet(), metavar="SYSGROUP",
                         help="include matrices marked from SYSGROUP")

def parse_options(args):
    from ._optparse import OptionGroup, OptionParser

    usage = "%prog [OPTION...] MATRIXDIR"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)

    add_optgroup_input_limiters(parser)

    with OptionGroup(parser, "Model parameters") as group:
        # XXX: change to --no-revcom, change default
        group.add_option("-r", "--revcom", action="store_true",
                         help="add reverse complement of all PWMs to model")
        group.add_option("-u", "--unbound", type=float,
                         default=DEFAULT_PROB_SILENT2UNBOUND,
                         metavar="PROB",
                         help="prior probability of unbound state"
                         " (default %s)" % DEFAULT_PROB_SILENT2UNBOUND)
        group.add_option("-p", "--priors",
                         metavar="FILE",
                         help="file containing prior probabilities of "
                         "unbound state and for bound states for different "
                         "transcription factors (supersedes --unbound)")
        group.add_option("-g", "--gc", type=float,
                         default=DEFAULT_GC_FRAC,
                         metavar="GC",
                         help=("G+C content between 0 and 1 (default %s)"
                               % DEFAULT_GC_FRAC))

    with OptionGroup(parser, "Cooperativity parameters") as group:
        group.add_option("-c", "--coop", action="store_true",
                         help="build model with cooperativity states")
        group.add_option("-d", "--coop-distance", type=float,
                         default=DEFAULT_COOP_DISTANCE,
                         help="Mean distance at which co-operativity is "
                         "effective; used to calculate P(coop->coop) "
                         "(default %s)" % DEFAULT_COOP_DISTANCE)
        group.add_option("-t", "--coop-strength", type=float,
                         default=DEFAULT_COOP_STRENGTH,
                         help="percentage increase in prior probability of "
                         "transitioning from an unbound cooperative state "
                         "to a bound state, compared to prior probability "
                         "of transitioning from unbound, non-cooperative "
                         "state to a bound state (range (0-1), default %s)"
                          % DEFAULT_COOP_STRENGTH)

    with OptionGroup(parser, "Output") as group:
        group.add_option("-o", "--output", default=OUTPUT_FILENAME,
                         metavar="FILE", help="output model to FILE "
                         "(default %s)" % OUTPUT_FILENAME)

    options, args = parser.parse_args(args)

    if len(args) != 1:
        parser.print_usage()
        sys.exit(1)

    if not 0 < options.coop_strength < 1:
        print "--coop-strength must be between 0 and 1 exclusive"
        parser.print_usage()
        sys.exit(1)

    if not 0 < options.gc < 1:
        print "--gc must be between 0 and 1 exclusive"
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    return pwm2sfl(output_filename=options.output,
                   prob_silent2unbound=options.unbound,
                   priors=options.priors,
                   coop=options.coop,
                   coop_distance=options.coop_distance,
                   coop_strength=options.coop_strength,
                   revcom=options.revcom, species_set=options.species,
                   gc_frac=options.gc,
                   sysgroup_set=options.sysgroup, *args)

if __name__ == "__main__":
    sys.exit(main())




#!/usr/bin/env python
from __future__ import division

__version__ = "$Revision: 413 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import re
import sys

from optparse import IndentedHelpFormatter
from pkg_resources import load_entry_point

from .._poly import substitute

class DriverHelpFormatter(IndentedHelpFormatter):
    """just replaces 'Options:' with 'Driver options:'"""
    def format_heading(self, heading):
        if heading == "Options":
            heading = "Driver options"

        return IndentedHelpFormatter.format_heading(self, heading)

class Driver(object):
    def __init__(self, command_args, colnames, measurement, write_seq,
                 start_default=None, limit_slice=slice(None)):
        assert not command_args

        self.colnames = colnames
        self.measurement = measurement
        self.write_seq = write_seq
        self.start_default = start_default

        limit_start = limit_slice.start
        if not limit_start:
            limit_start = 0
        self.limit_start = limit_start

    def __exit__(self, *exc_info):
        pass

    @classmethod
    def print_help(cls):
        return cls.get_option_parser().print_help()

    @classmethod
    def print_usage(cls):
        return cls.get_option_parser().print_usage()

    def defline_start(self, defline):
        if self.start_default is None:
            start = get_defline_start(defline)
        else:
            start = self.start_default

        limit_start = self.limit_start
        if limit_start is not None:
            start += limit_start

        return start

class FileDriver(Driver):
    def __init__(self, command_args, *args, **kwargs):
        if len(command_args) != 1:
            self.print_help()
            sys.exit(1)

        # set special characters on Poly distributed systems
        filename = substitute(command_args[0])

        self.filename = filename

        Driver.__init__(self, [], *args, **kwargs)

    @staticmethod
    def get_option_parser():
        from optparse import OptionParser

        usage = ("%prog --driver=DRIVER [OPTION...] MODEL SEQFILE "
                 "[DRIVEROPTION...] OUTFILE")
        return OptionParser(usage=usage, formatter=DriverHelpFormatter())

def load(name):
    return load_entry_point("sunflower", "sunflower.drivers", name)

# genomic format:
# chromosome:NCBI36:21:1:46944323:1 chromosome 21
# or
# chromosome:NCBI36:21:1:46944323:1:subseq(0,30000) chromosome 21

def defline_identifier(defline):
    first_word = defline.lstrip().partition(" ")[0]
    col_split = first_word.split(":")

    if len(col_split) >= 3:
        return ":".join(col_split[:3])
    else:
        return first_word

re_directive = re.compile(r"\[sunflower ([^]]+)\]")
def get_defline_start(defline):
    m_directive = re_directive.search(defline)
    if m_directive:
        texts = m_directive.group(1).split()
        directives = dict(text.split("=") for text in texts)

        return int(directives.get("start", 0))

    res = 0

    first_word = defline.partition(" ")[0]
    col_split = first_word.split(":")

    if len(col_split) >= 5:
        res = int(col_split[3])
    if len(col_split) >= 7 and col_split[-1].startswith("subseq("):
        res += int(col_split[-1][7:].partition(",")[0])

    return res

def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

__version__ = "$Revision: 413 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import cPickle as pickle
from os import extsep
import re
import sys

from numpy import NINF, allclose, array, empty, exp as aexp, intp
from pkg_resources import resource_stream

SPECIES_RESOURCENAME = "species.pck"

# these attributes are inherently part of every HMM we produce
ALPHABET = "ACGT"
ALPHABET_INT = range(len(ALPHABET))
LEN_ALPHABET = len(ALPHABET)

SPECIAL_STATE_NAMES = ["silent", "unbound"]
OFFSET_SILENT = 0
OFFSET_UNBOUND = 1

# suffixes that mark the beginning of a TF
TF_START_SUFFIXES = set(["0", "f0", "r0"])

# version of pickle being used
PICKLE_PROTOCOL = 2

EXT = "sfl"

# XXX: this stuff should be OO after conversion to HDF5

def get_params(model):
    a = model["logprobs_transition"]
    e = model["logprobs_emission"]
    cf = intp(model["connections_f"])
    cb = intp(model["connections_b"])

    return (a, e, cf, cb)

def check(model):
    """verify that appropriate probabilities sum to 0 or 1"""
    a = aexp(model["logprobs_transition"])
    e = aexp(model["logprobs_emission"])

    assert e[OFFSET_SILENT].sum() == 0.0

    # XXX: need to think carefully about how much tolerance to allow
    # here
    assert allclose(e[OFFSET_UNBOUND:].sum(-1), 1.0)

    assert allclose(a.sum(-1), 1.0)

def adjust(model, logprobs_emission):
    model["logprobs_emission"][OFFSET_UNBOUND] = logprobs_emission

def _load_from_resource(resourcename):
    full_resourcename = extsep.join([resourcename, EXT])

    return pickle.load(resource_stream("sunflower.data", full_resourcename))

def load(filename, resource=False):
    if resource:
        # first check the index
        species_info_resource = \
            resource_stream("sunflower.data", SPECIES_RESOURCENAME)
        species_infos = pickle.load(species_info_resource)

        try:
            resourcename, logprobs_emission = species_infos[filename]

            res = _load_from_resource(resourcename)
            adjust(res, logprobs_emission)
            return res
        except KeyError:
            # if that doesn't work, use the raw resource name
            return _load_from_resource(filename)
    else:
        return pickle.load(open(filename, "rb"))

def save(filename, model):
    if __debug__:
        check(model)

    with open(filename, "wb") as outfile:
        pickle.dump(model, outfile, PICKLE_PROTOCOL)

def fill_array(scalar, shape, dtype=None, *args, **kwargs):
    if dtype is None:
        dtype = array(scalar).dtype

    res = empty(shape, dtype, *args, **kwargs)
    res.fill(scalar)

    return res

def _arrayize_connection_list(connection_list, num_states):
    max_len = max(len(ks) for ks in connection_list) + 1

    res = fill_array(-1, (num_states, max_len), dtype=intp)

    for l, ks in enumerate(connection_list):
        res[l, 0] = len(ks)
        res[l, 1:len(ks)+1] = ks

    return res

def _make_connections(a, num_states, all_states):
    connection_list_f = [[] for _ in all_states]
    connection_list_b = [[] for _ in all_states]

    for k in all_states:
        for l in all_states:
            if a[k, l] != NINF:
                connection_list_f[l].append(k)
                connection_list_b[k].append(l)

    return [_arrayize_connection_list(connection_list, num_states)
            for connection_list in [connection_list_f, connection_list_b]]

def optimize(model):
    """
    add denormalized data

    The normalized data is really only (alphabet, desc, state_names,
    logprobs_transition, logprobs_emission)

    emitting_states is denormal unless the code in simulator has been
    updated to support more than one silent state.
    """

    num_states = len(model["state_names"])
    logprobs_a = model["logprobs_transition"]

    # not xrange because will be pickled
    all_states = range(num_states)
    emitting_states = range(OFFSET_UNBOUND, num_states)

    connections = _make_connections(logprobs_a, num_states, all_states)

    model.update(all_states=all_states,
                 emitting_states=emitting_states,
                 num_states=num_states,
                 connections_f=connections[0],
                 connections_b=connections[1])

def dump(model):
    print model["state_names"]
    print

    print "a = "
    print aexp(model["logprobs_transition"])
    print

    print "e = "
    print aexp(model["logprobs_emission"])
    print

    print "cf = %s" % model["connections_f"]
    print

    print "cb = %s" % model["connections_b"]

def main(args):
    for arg in args:
        dump(load(arg))

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
from __future__ import division

__version__ = "$Revision: 457 $"

# Copyright 2007, 2009 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import sys

from . import FileDriver, defline_identifier

class TabDelimDriver(FileDriver):
    def __enter__(self):
        outfile = open(self.filename, "w")
        self.file = outfile

        return self

    def __exit__(self, *exc_info):
        self.file.close()

    def __call__(self, arr, (defline, seq)):
        name = defline_identifier(defline)

        self.file.write(name)
        self.file.write("\t")
        self.file.write(str(arr))
        self.file.write("\n")

def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division

"""
_poly: provides pass-through substitute functions for Poly

chunk() ordinarily gets a work chunk for a distributed system
substitute() substitutes special identifiers in filename
"""

__version__ = "$Revision: 49 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import sys

def _chunk(collection):
    return collection

def _substitute(text):
    return text

try:
    from poly import chunk, substitute
except ImportError:
    chunk = _chunk
    substitute = _substitute

def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

"""Sunflower: model transcription factor binding to DNA

Sunflower models the simultaneous binding of transcription factors to
DNA. It uses a hidden Markov model that resembles a sunflower.
"""

__version__ = "1.1.0"

# Copyright 2007-2009 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import sys

# required for from __future__ import division, with_statement
assert sys.version_info >= (2, 5, 1)

import os
from subprocess import check_call

from ez_setup import use_setuptools
use_setuptools()

from setuptools import Extension, find_packages, setup
from setuptools.command.install import install

# provide a missing setuptools feature
class NumpyExtension(Extension):
    def __init__(self, *args, **kwargs):
        Extension.__init__(self, *args, **kwargs)

        self._include_dirs = self.include_dirs
        del self.include_dirs # restore overwritten property

    # warning: Extension is a classic class so it's not really read-only
    @property
    def include_dirs(self):
        from numpy import get_include

        return self._include_dirs + [get_include()]

doclines = __doc__.splitlines()
name, short_description = doclines[0].split(": ")
long_description = "\n".join(doclines[2:])

url = "http://www.ebi.ac.uk/~hoffman/software/%s/" % name.lower()
download_url = "%s%s-%s.tar.gz" % (url, name, __version__)

classifiers = ["Development Status :: 2 - Pre-Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: Other/Proprietary License",
               "Natural Language :: English",
               "Operating System :: POSIX :: Linux",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering :: Bio-Informatics"]

# XXX: sql driver should have [sql] at the end so it is only loaded sometimes
entry_points = """
[console_scripts]
h5attr = sunflower.h5attr:main
h5cat = sunflower.h5cat:main
pwm2sfl = sunflower.pwm2sfl:main
sunrecompose = sunflower.recompose:main
sunflower = sunflower.simulator:main
sunreport = sunflower.report:main
sunsplit = sunflower.split:main

[sunflower.drivers]
DEFAULT = sunflower.driver.hdf5:HDF5Driver
hdf5 = sunflower.driver.hdf5:HDF5Driver
sql = sunflower.driver.sql:SQLDriver
tabdelim = sunflower.driver.tabdelim:TabDelimDriver
"""

ext_modules = [NumpyExtension("_fwd_bwd",
                              ["src/_fwd_bwd.pyx", "src/fwd_bwd.c"])]

setup_requires = ["numpy>=1.0.3"]
install_requires = ["path>=2.2", "tables>=2.0.dev",
                    "textinput>=0.1.1", "setuptools>=0.6c9"] + setup_requires

extras_require = dict(distributed=["Poly>=0.1.1"],
                      develop=["Pyrex>=0.9.5.1a"],
                      sql=["docsql"])

# XXX: remove these when the upstream packages are updated to fix these issues
dependency_links = ["http://pypi.python.org/packages/source/p/path.py/path-2.2.zip"]

# XXX: test new setuptools does not require this?
# this is a workaround of the fact that setuptools doesn't install
# things that are in setup_requires even if they are also in
# install_requires
class SubprocessEasyInstall(install):
    def run(self):
        install.run(self)

        args = [sys.executable, "-m", "easy_install"] + setup_requires
        check_call(args)

if __name__ == "__main__":
    setup(name=name,
          version=__version__,
          description=short_description,
          author="Michael Hoffman",
          author_email="hoffman+%s@ebi.ac.uk" % name.lower(),
          url=url,
          download_url=download_url,
          license="all rights reserved",
          classifiers=classifiers,
          long_description=long_description,
          dependency_links=dependency_links,
          setup_requires=setup_requires,
          install_requires=install_requires,
          extras_require=extras_require,
          zip_safe=False,

          # XXX: this should be based off of __file__ instead
          packages=find_packages("."),
          ext_package="sunflower",
          ext_modules=ext_modules,
          include_package_data=True,
          entry_points=entry_points,
    #      cmdclass=dict(install=SubprocessEasyInstall)
          )



#!/usr/bin/env python
from __future__ import division, with_statement

"""
factorizers: break arrays up into factors
"""

__version__ = "$Revision: 457 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import csv
import sys

from numpy import (arange, array, empty, identity, isfinite, isnan, isneginf,
                   newaxis, ones, zeros)
import tabdelim # register unix-tab

from ._utils import FastaIterator, enum, progress
from .hmm import ALPHABET, ALPHABET_INT, LEN_ALPHABET

FACTOR_NAME_DEFAULT = "all"

LABELS_ALN = ["observed", "unobserved",
              "ambiguous", "insertion", "unaligned", "error"]
LABELS_CONS = ["conserved", "substitution"] + LABELS_ALN[2:]
LABELS_SUBS_ONLY = ["substitution/observed", "substitution/unobserved"]
LABELS_ALIGNED_ONLY = ["observed", "aligned"]

LEN_LABELS_ALN = len(LABELS_ALN)
LEN_LABELS_SUBS_ONLY = len(LABELS_SUBS_ONLY)
LEN_LABELS_ALIGNED_ONLY = len(LABELS_ALIGNED_ONLY)

LABELS_ALN_ENUM = enum(LABELS_ALN)
LABEL_ALN_OBS = LABELS_ALN_ENUM["observed"]
LABEL_ALN_UNOBS = LABELS_ALN_ENUM["unobserved"]
LABEL_ALN_UNALIGNED = LABELS_ALN_ENUM["unaligned"]

LABELS_SUBS_ONLY_ENUM = enum(LABELS_SUBS_ONLY)
LABEL_SUB_OBS = LABELS_SUBS_ONLY_ENUM["substitution/observed"]
LABEL_SUB_UNOBS = LABELS_SUBS_ONLY_ENUM["substitution/unobserved"]

LABELS_ALIGNED_ONLY_ENUM = enum(LABELS_ALIGNED_ONLY)
LABEL_ALIGNED_ONLY_OBS = LABELS_ALIGNED_ONLY_ENUM["observed"]
LABEL_ALIGNED_ONLY_ALIGNED = LABELS_ALIGNED_ONLY_ENUM["aligned"]

GAPPED_ALPHABET = ALPHABET + "N-.!"
LEN_GAPPED_ALPHABET = len(GAPPED_ALPHABET)
GAPPED_ALPHABET_ENUM = enum(GAPPED_ALPHABET)

GAPPED_ALPHABET2LABELS_ALN = [("N", "ambiguous"), ("-", "insertion"),
                              (".", "unaligned"), ("!", "error")]

IDENTITY_ALPHABET = identity(LEN_ALPHABET, bool)
SLICE_ALPHABET = slice(0, LEN_ALPHABET)

def gapped_seq2int(seq):
    res = [GAPPED_ALPHABET_ENUM[char] for char in seq.upper()]
    return array(res)

def load_seqs(filename):
    progress("loading sequences from %s..." % filename)
    with open(filename) as seqfile:
        iterator = FastaIterator(seqfile)
        res = dict((defline.partition(" ")[0], seq)
                   for defline, seq in iterator)

        progress("  done")
        return res

class Factorizer(object):
    """
    abstract base class: add an outer dimension to transcript_array
    for a factor
    """

    labels = NotImplemented

    def __getitem__(self, index):
        return self.labels[index]

    def __iter__(self):
        return iter(self.labels)

    def __call__(self, inarray):
        raise NotImplementedError

    def _init_res(self, inarray):
        inarray_shape = inarray.shape
        shape = (len(self.labels),) + inarray_shape
        dtype = inarray.atom.dtype

        return zeros(shape, dtype)

    def factor_arrays(self, inarrays):
        for inarray_index, inarray in enumerate(inarrays):
            name = inarray.name

            if not inarray_index % 10:
                progress(" %5d: %s", inarray_index, name)

            yield name, self(inarray), inarray.attrs


class TabdelimFactorizer(Factorizer):
    def __init__(self, filename):
        with open(filename) as tabfile:
            reader = csv.reader(tabfile, dialect="unix-tab")
            name2label = dict(reader)

        labels = list(sorted(set(name2label.itervalues())))
        # labels += "error"
        factor_enum = enum(labels)

        self.labels = labels
        self.name2factor = dict((name, factor_enum[label])
                                for name, label in name2label.iteritems())

    def __call__(self, inarray):
        res = self._init_res(inarray)

        factor = self.name2factor[inarray.name]
        res[factor] = inarray.read()

        return res


class AlignmentFactorizer(Factorizer):
    def __init__(self, filename):
        self.seqs = load_seqs(filename)

    def _init_res(self, inarray):
        res = Factorizer._init_res(self, inarray)

        try:
            seq = self.seqs[inarray.name]
        except KeyError:
            return res, None

        if res.shape[1] != len(seq):
            raise ValueError("length of alignment does not match that of sequence")

        seq_array = gapped_seq2int(seq)

        return res, seq_array


class ByNucleotideAlignmentFactorizer(AlignmentFactorizer):
    labels = LABELS_ALN

    def __init__(self, *args, **kwargs):
        AlignmentFactorizer.__init__(self, *args, **kwargs)

        mask = zeros((LEN_GAPPED_ALPHABET, LEN_LABELS_ALN, LEN_ALPHABET), bool)

        for gapped_letter, label_aln in GAPPED_ALPHABET2LABELS_ALN:
            gapped_alphabet_id = GAPPED_ALPHABET_ENUM[gapped_letter]
            label_aln_id = LABELS_ALN_ENUM[label_aln]

            mask[gapped_alphabet_id, label_aln_id] = True

        mask[SLICE_ALPHABET, LABEL_ALN_OBS] = IDENTITY_ALPHABET
        mask[SLICE_ALPHABET, LABEL_ALN_UNOBS] = ~IDENTITY_ALPHABET

        self.mask = mask

    def __call__(self, inarray):
        res, seq_array = self._init_res(inarray)

        if seq_array is None:
            res[LABEL_ALN_UNALIGNED] = inarray.read()
            return res

        where = self.mask[seq_array].transpose(1, 0, 2)
        res[where] = array([inarray.read()] * LEN_LABELS_ALN)[where]

        return res


class AlignedOnlyAlignmentFactorizer(AlignmentFactorizer):
    labels = LABELS_ALIGNED_ONLY

    def __init__(self, *args, **kwargs):
        AlignmentFactorizer.__init__(self, *args, **kwargs)

        shape = (LEN_GAPPED_ALPHABET, LEN_LABELS_ALIGNED_ONLY, LEN_ALPHABET)
        mask = zeros(shape, bool)

        mask[SLICE_ALPHABET, LABEL_ALIGNED_ONLY_OBS] = IDENTITY_ALPHABET
        mask[SLICE_ALPHABET, LABEL_ALIGNED_ONLY_ALIGNED] = True

        self.mask = mask

    def __call__(self, inarray):
        res, seq_array = self._init_res(inarray)

        if seq_array is None:
            return res

        where = self.mask[seq_array].transpose(1, 0, 2)
        res[where] = array([inarray.read()] * LEN_LABELS_ALIGNED_ONLY)[where]

        return res

def make_mask_bypos_alignment():
    # this is a different kind of mask
    # transforms a pair into something from LABELS_CONS

    # default: substitution
    res = ones((LEN_GAPPED_ALPHABET, LEN_GAPPED_ALPHABET), int)

    for nucleotide in ALPHABET_INT:
        res[nucleotide, nucleotide] = LABEL_ALN_OBS

    for gapped_letter, label_aln in GAPPED_ALPHABET2LABELS_ALN:
        gapped_alphabet_id = GAPPED_ALPHABET_ENUM[gapped_letter]
        label_cons_id = LABELS_ALN_ENUM[label_aln]

        res[gapped_alphabet_id, :] = label_cons_id
        res[:, gapped_alphabet_id] = label_cons_id

    return res

class ByPosAlignmentFactorizer(AlignmentFactorizer):
    labels = LABELS_CONS

    def __init__(self, *args, **kwargs):
        AlignmentFactorizer.__init__(self, *args, **kwargs)

        self.mask = make_mask_bypos_alignment()

    def __call__(self, inarray):
        res, seq_array = self._init_res(inarray)

        if seq_array is None:
            res[LABEL_ALN_UNALIGNED] = inarray.read()
            return res

        ref_array = gapped_seq2int(inarray.attrs.seq)

        where = self.mask[ref_array, seq_array]
        res[where, arange(inarray.shape[0])] = inarray.read()

        return res


class SubsOnlyAlignmentFactorizer(AlignmentFactorizer):
    labels = LABELS_SUBS_ONLY

    def __init__(self, *args, **kwargs):
        AlignmentFactorizer.__init__(self, *args, **kwargs)

        # (ref_seq, align_seq, label, mutation)
        shape = (LEN_GAPPED_ALPHABET, LEN_GAPPED_ALPHABET,
                 LEN_LABELS_SUBS_ONLY, LEN_ALPHABET)
        mask = zeros(shape, bool)

        for ref_nucleotide in ALPHABET_INT:
            for align_nucleotide in ALPHABET_INT:
                if ref_nucleotide == align_nucleotide:
                    continue

                submask = mask[ref_nucleotide, align_nucleotide]

                for mutation_nucleotide in ALPHABET_INT:
                    if mutation_nucleotide == align_nucleotide:
                        submask[LABEL_SUB_OBS, mutation_nucleotide] = True
                    else:
                        submask[LABEL_SUB_UNOBS, mutation_nucleotide] = True

        self.mask = mask

    def __call__(self, inarray):
        res, seq_array = self._init_res(inarray)
        ref_array = gapped_seq2int(inarray.attrs.seq)

        where = self.mask[ref_array, seq_array].transpose(1, 0, 2)
        res[where] = array([inarray.read()] * LEN_LABELS_SUBS_ONLY)[where]

        return res


class NullFactorizer(Factorizer):
    def __init__(self, name=FACTOR_NAME_DEFAULT):
        self.labels = [name]

    def __call__(self, inarray):
        res = self._init_res(inarray)
        res[0] = inarray.read()

        return res

# this will deal with zeros correctly as long as you do sums and
# divisors separately
def calc_total_datacol(addend_array, methodname="sum"):
    return getattr(addend_array, methodname)(-1)[..., newaxis]

def add_total(addend_array, methodname="sum"):
    "add an additional row at the end for total"
    orig_shape = addend_array.shape
    shape = list(orig_shape)
    shape[-1] = 1

    res = empty(shape, addend_array.dtype)
    res[..., :orig_shape[-1]] = addend_array
    res[..., -1] = getattr(addend_array, methodname)(-1)

    return res

def fmt_float(number):
    """
    NaN -> "NA"
    inf -> "Inf"
    """
    number = float(number)

    if isfinite(number):
        return str(number)

    if isnan(number):
        return "NA"

    if isneginf(number):
        return "-Inf"

    return "Inf"

def choose_factorizer(options):
    # XXX: allow stacking of factors
    opt_alignment = options.alignment
    if opt_alignment:
        return options.alignment_factorizer(opt_alignment)
    elif options.tab_delim:
        return TabdelimFactorizer(options.tab_delim)
    else:
        return NullFactorizer()

def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())

if __debug__:
    __pychecker__ = "no-returnvalues unusednames=tabdelim"



#!/usr/bin/env python
from __future__ import division

__version__ = "$Revision: 320 $"

from functools import wraps
from itertools import count
import platform
import sys

import pkg_resources

_DESC_TEMPLATE = "%s (%s on %s: %s Python/%s)"
def _init_desc(template=_DESC_TEMPLATE):
    node = platform.node()
    platform_str = platform.platform()
    python_version = platform.python_version()

    package_name = __name__.partition(".")[0]
    package = pkg_resources.require(package_name)[0]
    package_version = package.version

    return template % (" ".join(sys.argv), package_version, node,
                       platform_str, python_version)

DESC = _init_desc()

def _correct_range_point(point, start):
    if point is None:
        return None
    else:
        return point - start

def correct_range(r, start):
    res_start = _correct_range_point(r.start, start)
    res_stop = _correct_range_point(r.stop, start)

    return slice(res_start, res_stop, r.step)

def enum(iterable, n=0):
    """
    returns dict such that seq[enum(seq)[name]] == name
    """
    return dict(zip(iterable, count(n)))

def die(message, returncode=1):
    print >>sys.stderr, message
    sys.exit(returncode)

def enabled(func):
    return func

def disabled_noop(func):
    @wraps(func)
    def noop(*args, **kwargs):
        pass

    return noop

def disabled_passthrough(func):
    @wraps(func)
    def passthrough(x):
        return x

    return passthrough

def passthrough(x):
    return x

def enable_if(condition, disabled=disabled_noop):
    if condition:
        return enabled
    else:
        return disabled

@enable_if(sys.stderr.isatty())
def progress(text, *args):
    print >>sys.stderr, text % tuple(args)

@enable_if(sys.stderr.isatty(), disabled_passthrough)
def progress_iter(it, describe_func=passthrough):
    for index, item in enumerate(it):
        progress("%d: %s", index, describe_func(item))

        yield item

class FastaIterator(object):
    def __init__(self, handle):
        self._handle = handle
        self._defline = None

    def __iter__(self):
        return self

    def next(self):
        lines = []
        defline_old = self._defline

        while 1:
            line = self._handle.readline()
            if not line:
                if not defline_old and not lines:
                    raise StopIteration
                if defline_old:
                    self._defline = None
                    break
            elif line[0] == '>':
                self._defline = line[1:].rstrip()
                if defline_old or lines:
                    break
                else:
                    defline_old = self._defline
            else:
                lines.append(line.rstrip())

        return defline_old, ''.join(lines)

def main(args):
    pass

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



__version__ = "$Revision: 36 $"

# Copyright 2006-2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from c_numpy cimport PyArray_DATA, PyArray_GETCONTIGUOUS, import_array, ndarray, npy_intp
from fwd_bwd cimport bwd_inner, fwd_inner, inner_funcptr

import_array()

# I use f to represent f or b
cdef double call_inner(inner_funcptr func, ndarray f, ndarray x, ndarray a, ndarray e, ndarray c, npy_intp f_row_start):
    # declarations
    cdef ndarray f_contiguous "f_contiguous", a_contiguous "a_contiguous",
    cdef ndarray e_contiguous "e_contiguous", x_contiguous "x_contiguous"
    cdef ndarray c_contiguous "c_contiguous"

    cdef double *f_data "f_data", *a_data "a_data", *e_data "e_data"
    cdef npy_intp *x_data "x_data", *c_data "c_data"
    cdef npy_intp f_rows "f_rows"
    cdef npy_intp a_cols "a_cols", e_cols "e_cols"
    cdef npy_intp f_cols "f_cols", c_cols "c_cols"

    # array data access: f
    f_contiguous = PyArray_GETCONTIGUOUS(f)
    f_rows = f_contiguous.dimensions[0]
    f_cols = f_contiguous.dimensions[1]
    f_data = <double *>PyArray_DATA(f_contiguous)

    # array data access: x
    x_contiguous = PyArray_GETCONTIGUOUS(x)
    x_data = <npy_intp *>PyArray_DATA(x_contiguous)

    # array data access: a
    a_contiguous = PyArray_GETCONTIGUOUS(a)
    a_cols = a_contiguous.dimensions[1]
    a_data = <double *>PyArray_DATA(a_contiguous)

    # array data access: e
    e_contiguous = PyArray_GETCONTIGUOUS(e)
    e_cols = e_contiguous.dimensions[1]
    e_data = <double *>PyArray_DATA(e_contiguous)

    # array data access: c
    c_contiguous = PyArray_GETCONTIGUOUS(c)
    c_cols = c_contiguous.dimensions[1]
    c_data = <npy_intp *>PyArray_DATA(c_contiguous)

    return func(f_data, x_data, a_data, e_data, c_data, f_rows, f_cols, a_cols, e_cols, c_cols, f_row_start)

def fwd(ndarray f, ndarray x, ndarray a, ndarray e, ndarray c, npy_intp f_row_start):
    return call_inner(fwd_inner, f, x, a, e, c, f_row_start)

def bwd(ndarray b, ndarray x, ndarray a, ndarray e, ndarray c, npy_intp b_row_start):
    return call_inner(bwd_inner, b, x, a, e, c, b_row_start)



#!/usr/bin/env python
from __future__ import division, with_statement

"""
coop: adds and parameterizes cooperativity states between 
transcription factor "petals"
"""

__version__ = "$Revision: 428 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

from collections import defaultdict
import sys

from numpy import NINF, array, float64, log as alog, exp as aexp
from math import log, exp

from ._utils import DESC
from .hmm import (SPECIAL_STATE_NAMES, OFFSET_UNBOUND, OFFSET_SILENT,
                  LEN_ALPHABET, TF_START_FINAL_SUFFIXES,
                  get_state_names, get_tf_info, get_transitions, 
                  adjust_transitions, get_emissions, adjust_emissions,
                  fill_array, optimize, load, save, dump)

def load_priors(filename):

    strength_priors = {}
    distance_priors = {}
    state_offsets = {}
    coop_state_names = list()

    ## read the file - must be in the form TF1 TF2 P(TF1->TF2) P(COOP->COOP)
    ## assumes symmetric cooperativity
    state_offset = 0
    with open(filename) as infile:
        for line in infile:
            tokens = line.split()
            if len(tokens) == 4:

                tf1 = tokens[0]
                tf2 = tokens[1]
                strength = float(tokens[2])
                distance = float(tokens[3])

                assert strength < 1
                assert distance < 1

                ## there are generally not going to be huge numbers of
                ## TFs involved so it's ok (for now) to duplicate the work 
                ## for self-coop (tf1 == tf2) states
                if tf1 not in strength_priors:
                    strength_priors[tf1] = {}
                    distance_priors[tf1] = {}
                    state_offsets[tf1] = {}

                if tf2 not in strength_priors:
                    strength_priors[tf2] = {}
                    distance_priors[tf2] = {}
                    state_offsets[tf2] = {}

                strength_priors[tf1][tf2] = strength
                strength_priors[tf2][tf1] = strength

                distance_priors[tf1][tf2] = distance
                distance_priors[tf2][tf1] = distance

                ## XXX: very important that there is no duplication in the
                ## input state pairs - not sure how best to check this
                if tf1 == tf2:
                    coop_state_names.append("COOP.%s.%s" % (tf1, tf1))
                    state_offsets[tf1][tf1] = state_offset
                    state_offset += 1

                else:
                    coop_state_names.append("COOP.%s.%s" % (tf1, tf2))
                    state_offsets[tf1][tf2] = state_offset
                    state_offset += 1

                    coop_state_names.append("COOP.%s.%s" % (tf2, tf1))
                    state_offsets[tf2][tf1] = state_offset
                    state_offset += 1
 
    return (strength_priors, distance_priors, coop_state_names, state_offsets)

def add_coop(model, filename):

    ## read the coop priors
    (strength_priors, distance_priors, coop_state_names, \
         coop_state_offsets) = load_priors(filename)

    ## get the tf information and matrices
    state_names = get_state_names(model)
    (tf_start_offsets, tf_end_offsets, revcom) = get_tf_info(model)
    logprobs_a_old = get_transitions(model)
    logprobs_e_old = get_emissions(model)

    ## map tf names to start offsets and end offsets
    ## (works for both fwdonly and revcom)
    tf_to_start_offsets = {}
    tf_to_end_offsets = {}
    for tf_start_offset,tf_end_offset in zip(tf_start_offsets,tf_end_offsets):
        
        last_dot = state_names[tf_start_offset].rindex(".")
        tf_name = state_names[tf_start_offset][:last_dot]

        if tf_name not in tf_to_start_offsets:
            tf_to_start_offsets[tf_name] = list()
            tf_to_end_offsets[tf_name] = list()

        tf_to_start_offsets[tf_name].append(tf_start_offset)
        tf_to_end_offsets[tf_name].append(tf_end_offset)

    ## append the coop state names
    num_states_orig = len(state_names)
    state_names.extend(coop_state_names)
    num_states = len(state_names)

    ## reshape the transition and emissions matrices
    ## XXX: RAM usage doubles here. logprobs_a needs to be a more efficient
    ## data structure since it is sparse
    ## XXX: use resize() here?
    logprobs_a = fill_array(NINF, (num_states, num_states))
    logprobs_e = fill_array(NINF, (num_states, LEN_ALPHABET))

    logprobs_a_old_shape = logprobs_a_old.shape
    logprobs_e_old_shape = logprobs_e_old.shape

    logprobs_a[:logprobs_a_old_shape[0], :logprobs_a_old_shape[1]] = \
        logprobs_a_old
    logprobs_e[:logprobs_e_old_shape[0], :logprobs_e_old_shape[1]] = \
        logprobs_e_old

    ## update transitions
    for tf1 in coop_state_offsets:

        ## transitions for tf1 -> coop.tf1.tf2, coop.tf1.tf2 -> self,
        ## coop.tf1.tf2 -> tf2
        total_tf1_out = 0
        for tf2 in coop_state_offsets[tf1]:

            coop_state_offset = coop_state_offsets[tf1][tf2] + num_states_orig
            strength = strength_priors[tf1][tf2]
            distance = distance_priors[tf1][tf2]

            ## tf1 -> coop.tf1.tf2
            for tf1_end_offset in tf_to_end_offsets[tf1]:
                logprobs_a[tf1_end_offset, coop_state_offset] = log(strength)

            total_tf1_out += strength

            ## coop.tf1.tf2 -> coop.tf1.tf2
            logprobs_a[coop_state_offset, coop_state_offset] = log(distance)

            ## coop.tf1.tf2 -> tf2
            for tf2_start_offset in tf_to_start_offsets[tf2]:
                
                if revcom:
                    logprobs_a[coop_state_offset, tf2_start_offset] = log((1 - distance) / 2)
                else:
                    logprobs_a[coop_state_offset, tf2_start_offset] = log(1 - distance)

        assert total_tf1_out < 1

        ## tf1 -> silent
        for tf1_end_offset in tf_to_end_offsets[tf1]:
            logprobs_a[tf1_end_offset, OFFSET_SILENT] = log(1 - total_tf1_out)

    ## update emissions - for coop it is the same as unbound
    for tf1 in coop_state_offsets:
        for tf2 in coop_state_offsets[tf1]:
            coop_state_offset = coop_state_offsets[tf1][tf2] + num_states_orig
            logprobs_e[coop_state_offset] = logprobs_e[OFFSET_UNBOUND]

    return (logprobs_a, logprobs_e)

def transition_debug(logprobs_a, tf_start_offsets, tf_end_offsets, \
                         coop_state_offsets, num_states_orig):

    ## DEBUG: print only interesting transitions
    logprobs_ae = aexp(logprobs_a)
    all_offsets = list()
    all_offsets.append(OFFSET_SILENT)
    all_offsets.append(OFFSET_UNBOUND)
    for start,end in zip(tf_start_offsets, tf_end_offsets):
        all_offsets.append(start)
        all_offsets.append(end)
    for tf1 in coop_state_offsets:
        for tf2 in coop_state_offsets:
            all_offsets.append(coop_state_offsets[tf1][tf2] + num_states_orig)

    print all_offsets

    for i in all_offsets:
        for j in all_offsets:
            print "%0.4f" % logprobs_ae[i,j],
        print

def coop(priors_filename, infilename, outfilename, resource=False):

    model = load(infilename, resource)
    (logprobs_a, logprobs_e) = add_coop(model, priors_filename)

    adjust_transitions(model, logprobs_a)
    adjust_emissions(model, logprobs_e)
    model["desc"] = "; ".join([model["desc"], DESC])

    optimize(model)
    save(outfilename, model)

def parse_options(args):
    from optparse import OptionParser

    usage = "%prog [OPTION]... COOP-PRIORS IN-SFL OUT-SFL"
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-R", "--resource", action="store_true",
                      help="get SFLFILE from a resource that comes with the "
                      "sunflower distribution instead of a file")

    options, args = parser.parse_args(args)

    if not len(args) == 3:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):

    options, args = parse_options(args)

    return coop(resource=options.resource, *args)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import division, with_statement

"""
h5attr: set KEY=VALUE for every leaf of FILE
"""

__version__ = "$Revision: 192 $"

# Copyright 2007 Michael M. Hoffman <hoffman+software@ebi.ac.uk>

import sys

from .hdf5 import h5open
from ._utils import progress

def h5attr(h5filenames, set_attrs):
    for h5filename in h5filenames:
        progress(h5filename)

        with h5open(h5filename, "r+") as h5file:
            for leaf in h5file.walkNodes(classname="Leaf"):
                for key, value in set_attrs.iteritems():
                    setattr(leaf.attrs, key, value)

def parse_options(args):
    from ._optparse import OptionParser

    usage = "%prog [OPTION]... FILE..."
    version = "%%prog %s" % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-s", "--set-str", action="update", dest="set_attrs",
                      default={},
                      metavar="KEY VALUE", help="set KEY=VALUE")
    parser.add_option("-i", "--set-int", action="update", type=int,
                      dest="set_attrs",
                      metavar="KEY VALUE", help="set KEY=VALUE")

    options, args = parser.parse_args(args)

    if not len(args) >= 1:
        parser.print_usage()
        sys.exit(1)

    return options, args

def main(args=sys.argv[1:]):
    options, args = parse_options(args)

    return h5attr(args, options.set_attrs)

if __name__ == "__main__":
    sys.exit(main())
